#!/usr/bin/env python

"""
A tool for plotting an algorithm's ``Trading History`` from
a file in s3 from running the backtester with
the save to file option enabled:

::

    run_backtest_and_plot_history.py -t SPY -f <SAVE_HISTORY_TO_THIS_FILE>
"""

import argparse
import pandas as pd
import analysis_engine.consts as consts
import analysis_engine.plot_trading_history as plot_trading_history
import analysis_engine.load_history_dataset as load_history
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name='view-history-in-s3')


def plot_history_from_s3():
    """plot_history_from_s3

    Run a derived algorithm with an algorithm config dictionary

    :param config_dict: algorithm config dictionary
    """

    log.debug('start - plot')

    parser = argparse.ArgumentParser(
        description=(
            'plot a local algorithm trading history file'))
    parser.add_argument(
        '-b',
        help=(
            'saved in this s3 bucket'),
        required=False,
        dest='s3_bucket')
    parser.add_argument(
        '-k',
        help=(
            'saved in this s3 key'),
        required=False,
        dest='history_json_file')
    parser.add_argument(
        '-d',
        help=(
            'debug'),
        required=False,
        dest='debug',
        action='store_true')
    args = parser.parse_args()

    s3_access_key = consts.S3_ACCESS_KEY
    s3_secret_key = consts.S3_SECRET_KEY
    s3_region_name = consts.S3_REGION_NAME
    s3_address = consts.S3_ADDRESS
    s3_secure = consts.S3_SECURE
    compress = True

    s3_bucket = (
        'bt-spy-williamsr-2018-12-05-22-44-50-714400')
    s3_key = (
        '-181.55_netgain_9818.45_netvalue_NEGATIVE_'
        '10000.0_startbalance_1710.95_endbalance_'
        '30_shares_270.25_close_3_buys_0_sells_'
        '1_minbuyinds_1_minsellinds_'
        '43.52_seconds_'
        'trade_history-SPY_williamsr_test_'
        '0.73_for_176_of_24000.json')

    debug = False

    if args.debug:
        debug = True

    load_res = load_history.load_history_dataset(
        s3_enabled=True,
        s3_key=s3_key,
        s3_address=s3_address,
        s3_bucket=s3_bucket,
        s3_access_key=s3_access_key,
        s3_secret_key=s3_secret_key,
        s3_region_name=s3_region_name,
        s3_secure=s3_secure,
        compress=compress)

    algo_config = load_res.get(
        'algo_config_dict',
        None)
    algo_name = load_res.get(
        'algo_name',
        None)
    tickers = load_res.get(
        'tickers',
        [
            'SPY',
        ])
    ticker = tickers[0]

    log.info(
        'found algo: {}'.format(
            algo_name))
    log.info(
        'config: {}'.format(
            consts.ppj(algo_config)))

    history_df = load_res[ticker]
    history_df['date'] = pd.to_datetime(
        history_df['date'])
    history_df['minute'] = pd.to_datetime(
        history_df['minute'])
    ticker = history_df['ticker'].iloc[0]

    log.info('plotting history')

    first_date = history_df['date'].iloc[0]
    end_date = history_df['date'].iloc[-1]
    title = (
        'Trading History {}\n'
        'Backtest dates from {} to {}'.format(
            ticker,
            first_date,
            end_date))
    use_xcol = 'date'
    use_as_date_format = '%d\n%b'
    use_minute = False
    if 'minute' in history_df:
        found_valid_minute = history_df['minute'].iloc[0]
        if found_valid_minute:
            use_minute = True

    if use_minute:
        use_xcol = 'minute'
        use_as_date_format = '%d %H:%M:%S\n%b'
    xlabel = 'Dates vs Algo values'
    ylabel = 'Algo values'
    df_filter = (history_df['close'] > 1.00)

    # set default hloc columns:
    blue = None
    green = None
    orange = None

    red = 'close'
    blue = 'balance'

    if debug:
        for i, r in history_df.iterrows():
            log.info('{} - {}'.format(
                r['minute'],
                r['close']))
    # end of debug

    show_plot = True
    if show_plot:
        plot_trading_history.plot_trading_history(
            title=title,
            df=history_df,
            red=red,
            blue=blue,
            green=green,
            orange=orange,
            date_col=use_xcol,
            date_format=use_as_date_format,
            xlabel=xlabel,
            ylabel=ylabel,
            df_filter=df_filter,
            show_plot=True,
            dropna_for_all=True)

# end of plot_history_from_s3


if __name__ == '__main__':
    plot_history_from_s3()



"""
Extract an IEX dataset from Redis (S3 support coming soon) and
load it into a ``pandas.DataFrame``

Supported environment variables:

::

    # verbose logging in this module
    export DEBUG_EXTRACT=1

    # verbose logging for just Redis operations in this module
    export DEBUG_REDIS_EXTRACT=1

    # verbose logging for just S3 operations in this module
    export DEBUG_S3_EXTRACT=1

    # to show debug, trace logging please export ``SHARED_LOG_CFG``
    # to a debug logger json file. To turn on debugging for this
    # library, you can export this variable to the repo's
    # included file with the command:
    export SHARED_LOG_CFG=/opt/sa/analysis_engine/log/debug-logging.json

"""

import copy
import analysis_engine.iex.consts as iex_consts
import analysis_engine.extract_utils as extract_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def extract_daily_dataset(
        work_dict,
        scrub_mode='sort-by-date'):
    """extract_daily_dataset

    Fetch the IEX daily data for a ticker and
    return it as a pandas Dataframe

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = work_dict.get('label', 'extract')
    df_type = iex_consts.DATAFEED_DAILY
    df_str = iex_consts.get_datafeed_str(df_type=df_type)
    req = copy.deepcopy(work_dict)

    if 'redis_key' not in work_dict:
        # see if it's get dataset dictionary
        if 'daily' in req:
            req['redis_key'] = req['daily']
            req['s3_key'] = req['daily']
    # end of support for the get dataset dictionary

    log.debug(
        '{} - {} - start'.format(
            label,
            df_str))

    return extract_utils.perform_extract(
        df_type=df_type,
        df_str=df_str,
        work_dict=req,
        scrub_mode=scrub_mode)
# end of extract_daily_dataset


def extract_minute_dataset(
        work_dict,
        scrub_mode='sort-by-date'):
    """extract_minute_dataset

    Fetch the IEX minute intraday data for a ticker and
    return it as a pandas Dataframe

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = work_dict.get('label', 'extract')
    df_type = iex_consts.DATAFEED_MINUTE
    df_str = iex_consts.get_datafeed_str(df_type=df_type)
    req = copy.deepcopy(work_dict)

    if 'redis_key' not in work_dict:
        # see if it's get dataset dictionary
        if 'minute' in work_dict:
            req['redis_key'] = req['minute']
            req['s3_key'] = req['minute']
    # end of support for the get dataset dictionary

    log.debug(
        f'{label} - {df_str} - start')

    return extract_utils.perform_extract(
        df_type=df_type,
        df_str=df_str,
        work_dict=req,
        scrub_mode=scrub_mode)
# end of extract_minute_dataset


def extract_quote_dataset(
        work_dict,
        scrub_mode='sort-by-date'):
    """extract_quote_dataset

    Fetch the IEX quote data for a ticker and
    return it as a pandas Dataframe

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = work_dict.get('label', 'extract')
    df_type = iex_consts.DATAFEED_QUOTE
    df_str = iex_consts.get_datafeed_str(df_type=df_type)
    req = copy.deepcopy(work_dict)

    if 'redis_key' not in work_dict:
        # see if it's get dataset dictionary
        if 'quote' in work_dict:
            req['redis_key'] = req['quote']
            req['s3_key'] = req['quote']
    # end of support for the get dataset dictionary

    log.debug(
        '{} - {} - start'.format(
            label,
            df_str))

    return extract_utils.perform_extract(
        df_type=df_type,
        df_str=df_str,
        work_dict=req,
        scrub_mode=scrub_mode)
# end of extract_quote_dataset


def extract_stats_dataset(
        work_dict,
        scrub_mode='sort-by-date'):
    """extract_stats_dataset

    Fetch the IEX statistics data for a ticker and
    return it as a pandas Dataframe

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = work_dict.get('label', 'extract')
    df_type = iex_consts.DATAFEED_STATS
    df_str = iex_consts.get_datafeed_str(df_type=df_type)
    req = copy.deepcopy(work_dict)

    if 'redis_key' not in work_dict:
        # see if it's get dataset dictionary
        if 'stats' in work_dict:
            req['redis_key'] = req['stats']
            req['s3_key'] = req['stats']
    # end of support for the get dataset dictionary

    log.debug(
        '{} - {} - start'.format(
            label,
            df_str))

    return extract_utils.perform_extract(
        df_type=df_type,
        df_str=df_str,
        work_dict=req,
        scrub_mode=scrub_mode)
# end of extract_stats_dataset


def extract_peers_dataset(
        work_dict,
        scrub_mode='sort-by-date'):
    """extract_peers_dataset

    Fetch the IEX peers data for a ticker and
    return it as a pandas Dataframe

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = work_dict.get('label', 'extract')
    df_type = iex_consts.DATAFEED_PEERS
    df_str = iex_consts.get_datafeed_str(df_type=df_type)
    req = copy.deepcopy(work_dict)

    if 'redis_key' not in work_dict:
        # see if it's get dataset dictionary
        if 'peers' in work_dict:
            req['redis_key'] = req['peers']
            req['s3_key'] = req['peers']
    # end of support for the get dataset dictionary

    log.debug(
        '{} - {} - start'.format(
            label,
            df_str))

    return extract_utils.perform_extract(
        df_type=df_type,
        df_str=df_str,
        work_dict=req,
        scrub_mode=scrub_mode)
# end of extract_peers_dataset


def extract_news_dataset(
        work_dict,
        scrub_mode='sort-by-date'):
    """extract_news_dataset

    Fetch the IEX news data for a ticker and
    return it as a pandas Dataframe

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = work_dict.get('label', 'extract')
    df_type = iex_consts.DATAFEED_NEWS
    df_str = iex_consts.get_datafeed_str(df_type=df_type)
    req = copy.deepcopy(work_dict)

    if 'redis_key' not in work_dict:
        # see if it's get dataset dictionary
        if 'news1' in work_dict:
            req['redis_key'] = req['news1']
            req['s3_key'] = req['news1']
    # end of support for the get dataset dictionary

    log.debug(
        '{} - {} - start'.format(
            label,
            df_str))

    return extract_utils.perform_extract(
        df_type=df_type,
        df_str=df_str,
        work_dict=req,
        scrub_mode=scrub_mode)
# end of extract_news_dataset


def extract_financials_dataset(
        work_dict,
        scrub_mode='sort-by-date'):
    """extract_financials_dataset

    Fetch the IEX financial data for a ticker and
    return it as a pandas Dataframe

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = work_dict.get('label', 'extract')
    df_type = iex_consts.DATAFEED_FINANCIALS
    df_str = iex_consts.get_datafeed_str(df_type=df_type)
    req = copy.deepcopy(work_dict)

    if 'redis_key' not in work_dict:
        # see if it's get dataset dictionary
        if 'financials' in work_dict:
            req['redis_key'] = req['financials']
            req['s3_key'] = req['financials']
    # end of support for the get dataset dictionary

    log.debug(
        '{} - {} - start'.format(
            label,
            df_str))

    return extract_utils.perform_extract(
        df_type=df_type,
        df_str=df_str,
        work_dict=req,
        scrub_mode=scrub_mode)
# end of extract_financials_dataset


def extract_earnings_dataset(
        work_dict,
        scrub_mode='sort-by-date'):
    """extract_earnings_dataset

    Fetch the IEX earnings data for a ticker and
    return it as a pandas Dataframe

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = work_dict.get('label', 'extract')
    df_type = iex_consts.DATAFEED_EARNINGS
    df_str = iex_consts.get_datafeed_str(df_type=df_type)
    req = copy.deepcopy(work_dict)

    if 'redis_key' not in work_dict:
        # see if it's get dataset dictionary
        if 'earnings' in work_dict:
            req['redis_key'] = req['earnings']
            req['s3_key'] = req['earnings']
    # end of support for the get dataset dictionary

    log.debug(
        '{} - {} - start'.format(
            label,
            df_str))

    return extract_utils.perform_extract(
        df_type=df_type,
        df_str=df_str,
        work_dict=req,
        scrub_mode=scrub_mode)
# end of extract_earnings_dataset


def extract_dividends_dataset(
        work_dict,
        scrub_mode='sort-by-date'):
    """extract_dividends_dataset

    Fetch the IEX dividends data for a ticker and
    return it as a pandas Dataframe

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = work_dict.get('label', 'extract')
    df_type = iex_consts.DATAFEED_DIVIDENDS
    df_str = iex_consts.get_datafeed_str(df_type=df_type)
    req = copy.deepcopy(work_dict)

    if 'redis_key' not in work_dict:
        # see if it's get dataset dictionary
        if 'dividends' in work_dict:
            req['redis_key'] = req['dividends']
            req['s3_key'] = req['dividends']
    # end of support for the get dataset dictionary

    log.debug(
        '{} - {} - start'.format(
            label,
            df_str))

    return extract_utils.perform_extract(
        df_type=df_type,
        df_str=df_str,
        work_dict=req,
        scrub_mode=scrub_mode)
# end of extract_dividends_dataset


def extract_company_dataset(
        work_dict,
        scrub_mode='NO_SORT'):
    """extract_company_dataset

    Fetch the IEX company data for a ticker and
    return it as a pandas Dataframe

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = work_dict.get('label', 'extract')
    df_type = iex_consts.DATAFEED_COMPANY
    df_str = iex_consts.get_datafeed_str(df_type=df_type)
    req = copy.deepcopy(work_dict)

    if 'redis_key' not in work_dict:
        # see if it's get dataset dictionary
        if 'company' in work_dict:
            req['redis_key'] = req['company']
            req['s3_key'] = req['company']
    # end of support for the get dataset dictionary

    log.debug(
        '{} - {} - start'.format(
            label,
            df_str))

    return extract_utils.perform_extract(
        df_type=df_type,
        df_str=df_str,
        work_dict=req,
        scrub_mode=scrub_mode)
# end of extract_company_dataset



"""
Get latest pricing from cached IEX pricing data
"""

import copy
import json
import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import analysis_engine.api_requests as api_requests
import analysis_engine.iex.consts as iex_consts
import analysis_engine.iex.extract_df_from_redis as iex_extract_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def get_pricing_on_date(
        ticker,
        date_str=None,
        label=None):
    """get_pricing_on_date

    Get the latest pricing data from the
    cached IEX data in redis. Use this to
    keep costs down!

    .. code-block:: python

        import analysis_engine.iex.get_pricing_on_date as iex_cache
        print(iex_cache.get_pricing_on_date('SPY'))
        print(iex_cache.get_pricing_on_date(
            ticker='SPY',
            date_str='2019-02-07'))

    :param ticker: ticker string
    :param date_str: optional - string date
        to pull data from redis. if ``None`` use
        today's date. format is
        ``ae_consts.COMMON_TICK_DATE_FORMAT``
    :param label: log label from tracking
    """

    ret_dict = {
        'status': ae_consts.NOT_SET,
        'pricing_type': None,
        'high': None,
        'low': None,
        'open': None,
        'close': None,
        'volume': None,
        'date': None,
        'minute': None,
        'average': None,
        'changeOverTime': None,
        'label': None,
        'marketAverage': None,
        'marketChangeOverTime': None,
        'marketClose': None,
        'marketHigh': None,
        'marketLow': None,
        'marketNotional': None,
        'marketNumberOfTrades': None,
        'marketOpen': None,
        'marketVolume': None,
        'notional': None,
        'numberOfTrades': None
    }

    use_date_str = None
    if date_str:
        use_date_str = (
            f'{ticker}_{date_str}')

    all_extract_reqs = api_requests.get_ds_dict(
        ticker=ticker,
        base_key=use_date_str,
        label=label)

    minute_key = all_extract_reqs['minute']
    daily_key = all_extract_reqs['daily']
    base_ex_req = {
        'ticker': ticker,
        's3_bucket': 'pricing',
        's3_key': minute_key,
        'redis_key': minute_key,
        's3_enabled': True,
        's3_access_key': ae_consts.S3_ACCESS_KEY,
        's3_secret_key': ae_consts.S3_SECRET_KEY,
        's3_region_name': ae_consts.S3_REGION_NAME,
        's3_address': ae_consts.S3_ADDRESS,
        's3_secure': ae_consts.S3_SECURE,
        'redis_address': ae_consts.REDIS_ADDRESS,
        'redis_password': ae_consts.REDIS_PASSWORD,
        'redis_db': ae_consts.REDIS_DB,
        'redis_expire': ae_consts.REDIS_EXPIRE,
        'redis_enabled': True,
        'fetch_mode': 'td',
        'analysis_type': None,
        'iex_datasets': [],
        'debug': False,
        'label': label,
        'celery_disabled': True
    }
    log.debug(
        f'{ticker} - minute={minute_key} daily={daily_key}')
    reqs = []
    minute_ex_req = copy.deepcopy(base_ex_req)
    minute_ex_req['ex_type'] = iex_consts.FETCH_MINUTE
    minute_ex_req['iex_datasets'] = [
        iex_consts.FETCH_MINUTE
    ]
    reqs.append(minute_ex_req)
    daily_ex_req = copy.deepcopy(base_ex_req)
    daily_ex_req['ex_type'] = iex_consts.FETCH_DAILY
    daily_ex_req['s3_key'] = daily_key
    daily_ex_req['redis_key'] = daily_key
    daily_ex_req['iex_datasets'] = [
        iex_consts.FETCH_DAILY
    ]
    reqs.append(daily_ex_req)
    try:
        for ex_req in reqs:
            iex_status = ae_consts.FAILED
            iex_df = None
            if ex_req['ex_type'] == iex_consts.FETCH_MINUTE:
                iex_status, iex_df = \
                    iex_extract_utils.extract_minute_dataset(
                        work_dict=ex_req)
            else:
                iex_status, iex_df = \
                    iex_extract_utils.extract_daily_dataset(
                        work_dict=ex_req)
            # end of extracting

            if ae_consts.is_df(df=iex_df):
                if 'date' in iex_df:
                    iex_df.sort_values(
                        by=[
                            'date'
                        ],
                        ascending=True)
                    ret_dict = json.loads(iex_df.iloc[-1].to_json())
                    if 'date' in ret_dict:
                        try:
                            ret_dict['date'] = ae_utils.epoch_to_dt(
                                epoch=int(ret_dict['date']/1000),
                                use_utc=False,
                                convert_to_est=False).strftime(
                                    ae_consts.COMMON_TICK_DATE_FORMAT)

                        except Exception as f:
                            log.critical(
                                f'failed converting {ret_dict} date to str '
                                f'with ex={f}')
                    if ex_req['ex_type'] == iex_consts.FETCH_MINUTE:
                        ret_dict['pricing_type'] = 'minute'
                        ret_dict['minute'] = ret_dict.get(
                            'date',
                            None)
                    else:
                        ret_dict['pricing_type'] = 'daily'
                    ret_dict['status'] = iex_status
                    return ret_dict
            # if a valid df then return it
    except Exception as e:
        log.critical(
            f'failed to get {ticker} iex minute data with ex={e}')
        ret_dict['status'] = ae_consts.ERR
    # end of try/ex to get latest pricing

    return ret_dict
# end of get_pricing_on_date



"""
Helper for extracting details from Celery task and
sending it to a slack webhook.

Supported environment variables:

::

    # slack webhook
    export SLACK_WEBHOOK=https://hooks.slack.com/services/

"""

import os
import json
import requests
import tabulate as tb
import analysis_engine.consts as ae_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def post_success(msg,
                 jupyter=False,
                 block=False,
                 full_width=False):
    """Post a SUCCESS message to slack

    :param msg: A string, list, or dict to send to slack
    """
    result = {'status': ae_consts.FAILED}
    if not os.getenv('SLACK_WEBHOOK', False):
        log.info(
            'post_success - please add a SLACK_WEBHOOK environment '
            'variable to publish messages')
        return result
    if msg:
        attachments = [{"attachments":
                       [{"color": "good", "title": "SUCCESS"}]}]
        fields = parse_msg(msg, block=block)
        if fields:
            if full_width:
                attachments.append({"text": fields[0].pop("value")})
            else:
                attachments[0]["attachments"][0]["fields"] = fields
            result = post(attachments, jupyter=jupyter)
    return result


def post_failure(msg,
                 jupyter=False,
                 block=False,
                 full_width=False):
    """Post a FAILURE message to slack

    :param msg: A string, list, or dict to send to slack
    """
    result = {'status': ae_consts.FAILED}
    if not os.getenv('SLACK_WEBHOOK', False):
        log.info(
            'post_failure - please add a SLACK_WEBHOOK environment '
            'variable to publish messages')
        return result
    if msg:
        attachments = [{"attachments":
                       [{"color": "danger", "title": "FAILED"}]}]
        fields = parse_msg(msg, block=block)
        if fields:
            if full_width:
                attachments.append({"text": fields[0].pop("value")})
            else:
                attachments[0]["attachments"][0]["fields"] = fields
            result = post(attachments, jupyter=jupyter)
    return result


def post_message(msg,
                 jupyter=False,
                 block=False,
                 full_width=False):
    """Post any message to slack

    :param msg: A string, list, or dict to send to slack
    """
    result = {'status': ae_consts.FAILED}
    if not os.getenv('SLACK_WEBHOOK', False):
        log.info(
            'post_message - please add a SLACK_WEBHOOK environment '
            'variable to publish messages')
        return result
    if not os.getenv('SLACK_WEBHOOK', False):
        log.info(
            'post_message - please add a SLACK_WEBHOOK environment '
            'variable for to work')
        return result
    if msg:
        attachments = [{"attachments": [{"title": "MESSAGE"}]}]
        fields = parse_msg(msg, block=block)
        if fields:
            if full_width:
                attachments.append({"text": fields[0].pop("value")})
            else:
                attachments[0]["attachments"][0]["fields"] = fields
            result = post(attachments, jupyter=jupyter)
    return result


def parse_msg(msg, block=False):
    """Create an array of fields for slack from the msg type

    :param msg: A string, list, or dict to massage for sending to slack
    """
    if type(msg) is str:
        if block:
            return [{"value": "```{}```".format(msg)}]
        return [{"value": msg}]
    elif type(msg) is list:
        if block:
            string_list = '\n'.join("{}".format(str(x)) for x in msg)
            return [{"value": "```{}```".format(string_list)}]
        return [{"value": str(x)} for x in msg]
    elif type(msg) is dict:
        if block:
            string_dict = '\n'.join(
                "{}: {}".format(str(k), str(v)) for k, v in msg.items())
            return [{"value": "```{}```".format(string_dict)}]
        return [{"value": "{}: {}".format(
            str(k), str(v))} for k, v in msg.items()]
    return None


def post(attachments, jupyter=False):
    """Send created attachments to slack

    :param attachments: Values to post to slack
    """
    SLACK_WEBHOOK = ae_consts.ev('SLACK_WEBHOOK', None)
    result = {'status': ae_consts.FAILED}
    if not os.getenv('SLACK_WEBHOOK', False):
        log.info(
            'post - please add a SLACK_WEBHOOK environment '
            'variable to publish messages')
        return result
    if attachments and SLACK_WEBHOOK:
        try:
            # if not jupyter:
            #     log.debug(('Attempting to post attachments={} '
            #               'to slack_webhook exists').format(attachments))
            for attachment in attachments:
                r = requests.post(SLACK_WEBHOOK, data=json.dumps(attachment))
                if str(r.status_code) == "200":
                    # log.info((
                    #   'Successful post of attachment={} '
                    #   'to slack_webhook').format(
                    #       attachment if not jupyter else
                    #       True if attachment else False))
                    result['status'] = ae_consts.SUCCESS
                else:
                    log.error(('Failed to post attachment={} '
                               'with status_code={}').format(
                                   attachment if not jupyter else
                                   True if attachment else False,
                                   r.status_code))
                    result['status'] = ae_consts.FAILED
                    break
        except Exception as e:
            log.error(('Failed to post attachments={} '
                       'with ex={}').format(
                           attachments if not jupyter else
                           True if attachments else False,
                           e))
            result['status'] = ae_consts.ERR
            result['err'] = e
    else:
        log.info(('Skipping post to slack due to missing '
                  'attachments={} or SLACK_WEBHOOK missing={}').format(
                      attachments if not jupyter else
                      True if attachments else False,
                      False if SLACK_WEBHOOK else True))
    return result


def post_df(
        df,
        columns=None,
        block=True,
        jupyter=True,
        full_width=True,
        tablefmt='github'):
    """post_df

    Post a ``pandas.DataFrame`` to Slack

    :param df: ``pandas.DataFrame`` object
    :param columns: ordered list of columns to for the table
                    header row
                    (``None`` by default)
    :param block: bool for
                  post as a Slack-formatted block ```like this```
                  (``True`` by default)
    :param jupyter: bool for
                    jupyter attachment handling
                    (``True`` by default)
    :param full_width: bool to ensure the width is preserved
                       the Slack message  (``True`` by default)
    :param tablefmt: string for table format (``github`` by default).
                     Additional format values can be found on:
                     https://bitbucket.org/astanin/python-tabulate
    """

    if not os.getenv('SLACK_WEBHOOK', False):
        log.info(
            'post_df - please add a SLACK_WEBHOOK environment '
            'variable to publish messages')
        return
    if not hasattr(df, 'index'):
        log.debug('post_df - no df ')
        return

    log.debug('post_df - df.index={} columns={} fmt={}'.format(
        len(df.index),
        columns,
        tablefmt))

    msg = None
    if columns:
        msg = tb.tabulate(
            df[columns],
            headers=columns,
            tablefmt=tablefmt)
    else:
        msg = tb.tabulate(
            df,
            tablefmt=tablefmt)
    # end of if/else

    post_success(
        msg=msg,
        block=block,
        jupyter=jupyter,
        full_width=full_width)
# end of post_df


def post_cb(
        msg,
        block=True,
        jupyter=True,
        full_width=True,
        tablefmt='github'):
    """post_cb

    Post a text messsage as a code block to Slack

    :param msg: text message (pre-formatting is not necessary)
    :param block: bool for
                  post as a Slack-formatted block ```like this```
                  (``True`` by default)
    :param jupyter: bool for
                    jupyter attachment handling
                    (``True`` by default)
    :param full_width: bool to ensure the width is preserved
                       the Slack message  (``True`` by default)
    :param tablefmt: string for table format (``github`` by default).
                     Additional format values can be found on:
                     https://bitbucket.org/astanin/python-tabulate
    """

    if not os.getenv('SLACK_WEBHOOK', False):
        log.info(
            'post_cb - please add a SLACK_WEBHOOK environment '
            'variable to publish messages')
        return
    if not msg:
        log.debug('post_cb - no msg')
        return

    log.debug('post_cb - msg={} fmt={}'.format(
        len(msg),
        tablefmt))

    post_success(
        msg=msg,
        block=block,
        jupyter=jupyter,
        full_width=full_width)
# end of post_cb


def post_plot(plot,
              filename='tmp',
              title=None):
    """post_plot

    Post a matlibplot plot to Slack

    :param plot: matlibplot pyplot figure
    :param filename: filename of plot
    :param title: Title for slack plot postiing
    """

    result = {'status': ae_consts.FAILED}
    if not os.getenv('SLACK_ACCESS_TOKEN', False):
        log.info(
            'post_plot - please add a SLACK_ACCESS_TOKEN environment '
            'variable to publish plots')
        return result
    if not filename:
        log.info(
            'post_plot - no filename provided'
        )
        return result
    url = 'https://slack.com/api/files.upload'
    filename = f'{filename.split(".")[0]}.png'
    channels = [f'#{channel}' for channel in os.getenv(
        'SLACK_PUBLISH_PLOT_CHANNELS',
        'general').split(',')]
    try:
        log.info(f'post_plot - temporarily saving plot: {filename}')
        plot.savefig(filename)
        tmp_file = {
            'file': (filename, open(filename, 'rb'), 'png')
        }
        payload = {
            'channels':        channels,
            'filename':        filename,
            'title':           title if title else filename,
            'token':           os.getenv('SLACK_ACCESS_TOKEN'),
        }
        log.info(f'post_plot - sending payload: {payload} to url: {url}')
        r = requests.post(url,
                          params=payload,
                          files=tmp_file)
        log.info(f'post_plot - removing temporarily saved plot: {filename}')
        os.remove(filename)
        if str(r.status_code) == "200":
            log.info(f'post_plot - posted plot to slack channels: {channels}')
            result['status'] = ae_consts.SUCCESS
    except Exception as e:
        log.info(f'post_plot - failed with Exception: {e}')
        result['status'] = ae_consts.ERR
        result['err'] = e
    return result
# end of post_plot



"""
**IEX Cloud Environment Variables**

.. code-block:: python

    IEX_API_VERSION = os.getenv(
        'IEX_API_VERSION',
        'beta')
    IEX_URL_BASE = os.getenv(
        'IEX_URL',
        f'https://cloud.iexapis.com/{IEX_API_VERSION}')
    IEX_URL_BASE_V1 = os.getenv(
        'IEX_URL_V1',
        'https://api.iextrading.com/1.0/')
    IEX_TOKEN = os.getenv(
        'IEX_TOKEN',
        None)
    IEX_PROXIES = os.getenv(
        'IEX_PROXIES',
        None)
    DEFAULT_FETCH_DATASETS="daily,minute,quote,stats,
    peers,news,financials,earnings,dividends,company"

"""

import os


FETCH_DAILY = 800
FETCH_MINUTE = 801
FETCH_QUOTE = 802
FETCH_STATS = 803
FETCH_PEERS = 804
FETCH_NEWS = 805
FETCH_FINANCIALS = 806
FETCH_EARNINGS = 807
FETCH_DIVIDENDS = 808
FETCH_COMPANY = 809

DATAFEED_DAILY = 900
DATAFEED_MINUTE = 901
DATAFEED_QUOTE = 902
DATAFEED_STATS = 903
DATAFEED_PEERS = 904
DATAFEED_NEWS = 905
DATAFEED_FINANCIALS = 906
DATAFEED_EARNINGS = 907
DATAFEED_DIVIDENDS = 908
DATAFEED_COMPANY = 909

DEFAULT_FETCH_DATASETS = [
    FETCH_DAILY,
    FETCH_MINUTE,
    FETCH_QUOTE,
    FETCH_STATS,
    FETCH_PEERS,
    FETCH_NEWS,
    FETCH_FINANCIALS,
    FETCH_EARNINGS,
    FETCH_DIVIDENDS,
    FETCH_COMPANY
]
TIMESENSITIVE_DATASETS = [
    FETCH_MINUTE,
    FETCH_QUOTE,
    FETCH_NEWS
]
FUNDAMENTAL_DATASETS = [
    FETCH_QUOTE,
    FETCH_FINANCIALS,
    FETCH_EARNINGS,
    FETCH_DIVIDENDS,
    FETCH_STATS
]

IEX_DATE_FORMAT = '%Y-%m-%d'
IEX_TICK_FORMAT = '%Y-%m-%d %H:%M:%S'
IEX_FETCH_MINUTE_FORMAT = '%H:%M'

# IEX Cloud Environment Variables
IEX_API_VERSION = os.getenv(
    'IEX_API_VERSION',
    'beta')
IEX_URL_BASE = os.getenv(
    'IEX_URL',
    f'https://cloud.iexapis.com/{IEX_API_VERSION}')
IEX_URL_BASE_V1 = os.getenv(
    'IEX_URL_V1',
    'https://api.iextrading.com/1.0/')
IEX_TOKEN = os.getenv(
    'IEX_TOKEN',
    None)
IEX_PROXIES = os.getenv(
    'IEX_PROXIES',
    None)
IEX_DATE_FIELDS = [
    'date',
    'EPSReportDate',
    'fiscalEndDate',
    'exDate',
    'declaredDate',
    'paymentDate',
    'recordDate',
    'reportDate',
    'datetime',
    'expectedDate',
    'latestTime',
    'DailyListTimestamp',
    'RecordUpdateTime']
IEX_TIME_FIELDS = [
    'closeTime',
    'close.time',
    'delayedPriceTime',
    'extendedPriceTime',
    'iexLastUpdated',
    'latestTime',
    'openTime',
    'open.time'
    'processedTime',
    'time',
    'timestamp',
    'lastUpdated']
IEX_EPOCH_FIELDS = [
    'datetime'
]
IEX_SECOND_FIELDS = []

ENV_FETCH_DATASETS = os.getenv(
    'DEFAULT_FETCH_DATASETS_IEX',
    None)
if ENV_FETCH_DATASETS:
    SPLIT_FETCH_DATASETS_IEX = ENV_FETCH_DATASETS.split(',')
    DEFAULT_FETCH_DATASETS = []
    for d in SPLIT_FETCH_DATASETS_IEX:
        if d == 'minute':
            DEFAULT_FETCH_DATASETS.append(
                FETCH_MINUTE)
        elif d == 'daily':
            DEFAULT_FETCH_DATASETS.append(
                FETCH_DAILY)
        elif d == 'quote':
            DEFAULT_FETCH_DATASETS.append(
                FETCH_QUOTE)
        elif d == 'stats':
            DEFAULT_FETCH_DATASETS.append(
                FETCH_STATS)
        elif d == 'peers':
            DEFAULT_FETCH_DATASETS.append(
                FETCH_PEERS)
        elif d == 'news':
            DEFAULT_FETCH_DATASETS.append(
                FETCH_NEWS)
        elif d == 'financials':
            DEFAULT_FETCH_DATASETS.append(
                FETCH_FINANCIALS)
        elif d == 'earnings':
            DEFAULT_FETCH_DATASETS.append(
                FETCH_EARNINGS)
        elif d == 'dividends':
            DEFAULT_FETCH_DATASETS.append(
                FETCH_DIVIDENDS)
        elif d == 'company':
            DEFAULT_FETCH_DATASETS.append(
                FETCH_COMPANY)
# end of building env-datasets to get

FETCH_DATASETS = DEFAULT_FETCH_DATASETS


def get_ft_str(
        ft_type):
    """get_ft_str

    :param ft_type: enum fetch type value to return
                    as a string
    """
    ft_str = str(ft_type).lower()
    if ft_type == FETCH_DAILY or ft_str == 'daily':
        return 'daily'
    elif ft_type == FETCH_MINUTE or ft_str == 'minute':
        return 'minute'
    elif ft_type == FETCH_QUOTE or ft_str == 'quote':
        return 'quote'
    elif ft_type == FETCH_STATS or ft_str == 'stats':
        return 'stats'
    elif ft_type == FETCH_PEERS or ft_str == 'peers':
        return 'peers'
    elif ft_type == FETCH_NEWS or ft_str == 'news':
        return 'news'
    elif ft_type == FETCH_FINANCIALS or ft_str == 'financials':
        return 'financials'
    elif ft_type == FETCH_EARNINGS or ft_str == 'earnings':
        return 'earnings'
    elif ft_type == FETCH_DIVIDENDS or ft_str == 'dividends':
        return 'dividends'
    elif ft_type == FETCH_COMPANY or ft_str == 'company':
        return 'company'
    else:
        return 'unsupported ft_type={}'.format(
            ft_type)
# end of get_ft_str


def get_datafeed_str(
        df_type):
    """get_datafeed_str

    :param df_type: enum fetch type value to return
                    as a string
    """
    if df_type == DATAFEED_DAILY:
        return 'daily'
    elif df_type == DATAFEED_MINUTE:
        return 'minute'
    elif df_type == DATAFEED_QUOTE:
        return 'quote'
    elif df_type == DATAFEED_STATS:
        return 'stats'
    elif df_type == DATAFEED_PEERS:
        return 'peers'
    elif df_type == DATAFEED_NEWS:
        return 'news'
    elif df_type == DATAFEED_FINANCIALS:
        return 'financials'
    elif df_type == DATAFEED_EARNINGS:
        return 'earnings'
    elif df_type == DATAFEED_DIVIDENDS:
        return 'dividends'
    elif df_type == DATAFEED_COMPANY:
        return 'company'
    else:
        return 'unsupported df_type={}'.format(
            df_type)
# end of get_datafeed_str



"""
Helper for determining the pricing for
an entry position on a bear put spread
"""

import analysis_engine.consts as ae_consts
import analysis_engine.build_option_spread_details as spread_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def build_entry_put_spread_details(
        ticker,
        close,
        num_contracts,
        low_strike,
        low_ask,
        low_bid,
        high_strike,
        high_ask,
        high_bid):
    """build_entry_put_spread_details

    Calculate pricing information for
    buying into ``Vertical Bear Put Option Spread`` contracts

    :param ticker: string ticker symbol
    :param num_contracts: integer number of contracts
    :param low_strike: float - strike for
        the low leg of the spread
    :param low_ask: float - ask price for
        the low leg of the spread
    :param low_bid: float - bid price for
        the low leg of the spread
    :param high_strike: float - strike  for
        the high leg of the spread
    :param high_ask: float - ask price for
        the high leg of the spread
    :param high_bid: float - bid price for
        the high leg of the spread
    """

    spread_details = spread_utils.build_option_spread_details(
        trade_type=ae_consts.TRADE_ENTRY,
        spread_type=ae_consts.SPREAD_VERTICAL_BEAR,
        option_type=ae_consts.OPTION_PUT,
        close=close,
        num_contracts=num_contracts,
        low_strike=low_strike,
        low_ask=low_ask,
        low_bid=low_bid,
        high_strike=high_strike,
        high_ask=high_ask,
        high_bid=high_bid)

    log.debug(
        '{} type={} spread={} option={} close={} spread={}'.format(
            ticker,
            ae_consts.get_status(status=spread_details['trade_type']),
            ae_consts.get_status(status=spread_details['spread_type']),
            ae_consts.get_status(status=spread_details['option_type']),
            close,
            ae_consts.ppj(spread_details)))

    return spread_details
# end of build_entry_put_spread_details



"""
Option Date Utilities
=====================

These are a collection of functions for determining
when the current options cycle expires (3rd Friday of most months)
and for calculating historical option expiration dates.

.. tip:: If you need to automate looking up the current option
    cycle expiration, then please checkout using the script:

    ::

        /opt/sa/analysis_engine/scripts/print_next_expiration_date.py
        2018-10-19

"""

import datetime
import pandas.tseries.offsets as pd_bday
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def get_options_for_years(
        years=[
            '2014',
            '2015',
            '2016',
            '2016',
            '2017',
            '2018',
            '2019',
            '2020',
            '2021',
            '2022',
        ]):
    """get_options_for_years

    :param years: number of years back
    :param months: number of months to build year
    """

    entities = []
    months = [
        '01',
        '02',
        '03',
        '04',
        '05',
        '06',
        '07',
        '08',
        '09',
        '10',
        '11',
        '12'
    ]
    option_expirations = {
    }
    opts = []

    for year in years:
        for month in months:
            target_date_str = str(month + '-01-' + year)
            target_date = datetime.datetime.strptime(
                target_date_str,
                '%m-%d-%Y')
            option_expiration_date = option_expiration(
                target_date)
            option_expiration_str = option_expiration_date.strftime(
                '%m-%d-%Y')
            option_expirations[option_expiration_str] = option_expiration_date
            opts.append(option_expiration_date.strftime(
                '%m-%d-%Y'))
        # end of for all months
    # end of building option expiration dates

    now = datetime.datetime.now() + datetime.timedelta(days=0)
    num_legs = 20
    num_done = 1
    last_exp_date = 0
    date_format = '%m-%d-%Y'
    str_output = now.strftime(date_format)
    log.info(
        'current date={}'.format(
            str_output))

    entities.append(str_output)

    for option_exp in opts:
        option_exp_date = datetime.datetime.strptime(
            option_exp,
            date_format)
        if option_exp_date >= now:
            if num_legs > 0:
                if (last_exp_date == 0):
                    delta = (option_exp_date - now).days
                else:
                    delta = (option_exp_date - last_exp_date).days

                entities.append(str(delta))
                entities.append(option_exp_date.strftime('%m-%d-%Y'))
                entities.append('Leg ' + str(num_done))

            else:
                break
            num_legs -= 1
            num_done += 1
            last_exp_date = option_exp_date
    # end of processing

    return entities
# end of get_options_for_years


def historical_options(
        years=[
            '2014',
            '2015',
            '2016',
            '2017',
            '2018',
            '2019',
            '2020',
            '2021',
            '2022',
            '2023',
            '2024',
            '2025',
            '2026',
            '2027',
            '2028'
        ]):
    """historical_options

    :param years: years to build
    """

    entities = []
    months = [
        '01',
        '02',
        '03',
        '04',
        '05',
        '06',
        '07',
        '08',
        '09',
        '10',
        '11',
        '12'
    ]
    option_expirations = {
    }
    opts = []

    for year in years:
        for month in months:
            target_date_str = str(month + '-01-' + year)
            target_date = datetime.datetime.strptime(
                target_date_str,
                '%m-%d-%Y')
            option_expiration_date = option_expiration(
                target_date)
            option_expiration_str = option_expiration_date.strftime(
                '%m-%d-%Y')
            option_expirations[option_expiration_str] = option_expiration_date
            opts.append(option_expiration_date.strftime(
                '%m-%d-%Y'))
        # end of for all months

    # end of building option expiration dates

    now = datetime.datetime.strptime('01-01-2009', '%m-%d-%Y')
    num_legs = 400
    num_done = 1
    date_format = '%m-%d-%Y'

    for option_exp in opts:
        option_exp_date = datetime.datetime.strptime(option_exp, date_format)
        if option_exp_date >= now:
            if num_legs > 0:
                entities.append(option_exp_date.strftime('%m-%d-%Y'))
            else:
                break
            num_legs -= 1
            num_done += 1
    # end of processing

    return entities
# end of historical_options


def get_options_between_dates(
        start_date,
        end_date):
    """get_options_between_dates

    :param start_date: start date
    :param end_date: end date
    """
    valid_options = []

    for rec in historical_options():
        opt_date = datetime.datetime.strptime(
            str(rec),
            '%m-%d-%Y').date()
        if start_date <= opt_date <= end_date:
            valid_options.append(opt_date.strftime('%Y-%m-%d'))

    return valid_options
# end of get_options_between_dates


def option_expiration(
        date=None):
    """option_expiration

    :param date: date to find the current expiration
    """
    cur_date = date
    if not cur_date:
        cur_date = datetime.datetime.now()
    while (not (cur_date.weekday() == 4 and 14 < cur_date.day < 22)):
        cur_date = cur_date + datetime.timedelta(days=1)
    return cur_date
# end of option_expiration


def get_options_for_today():
    """get_options_for_today

    Get a list of option expiration nodes where the last cell
    has the current option cycle's expiration date.

    """
    cur_date = datetime.datetime.now()
    cycle_exps = historical_options()
    previous_exp = None
    valid_option_exps = []
    for idx, org_exp_date_str in enumerate(cycle_exps):
        log.debug(
            'cycle={} expiration={}'.format(
                idx,
                org_exp_date_str))
        exp_date = datetime.datetime.strptime(
            org_exp_date_str,
            '%m-%d-%Y')
        exp_date_str = exp_date.strftime(
            '%Y-%m-%d')

        cycle_start_date = exp_date - pd_bday.BDay(19)
        if previous_exp:
            cycle_start_date = previous_exp + pd_bday.BDay(1)
        cycle_start_date_str = cycle_start_date.strftime(
            '%m-%d-%Y')
        valid_option_exps.append({
            'exp_date': exp_date,
            'exp_date_str': exp_date_str,
            'cycle_start': cycle_start_date,
            'cycle_start_str': cycle_start_date_str
        })
        if cur_date < exp_date:
            break
        previous_exp = exp_date
    # end of for all historical options

    return valid_option_exps
# end of get_options_for_today



"""
Extract an TD dataset from Redis (S3 support coming soon) and
load it into a ``pandas.DataFrame``

Supported environment variables:

::

    # verbose logging in this module
    export DEBUG_EXTRACT=1

    # verbose logging for just Redis operations in this module
    export DEBUG_REDIS_EXTRACT=1

    # verbose logging for just S3 operations in this module
    export DEBUG_S3_EXTRACT=1

    # to show debug, trace logging please export ``SHARED_LOG_CFG``
    # to a debug logger json file. To turn on debugging for this
    # library, you can export this variable to the repo's
    # included file with the command:
    export SHARED_LOG_CFG=/opt/sa/analysis_engine/log/debug-logging.json

"""

import pandas as pd
import analysis_engine.consts as ae_consts
import analysis_engine.dataset_scrub_utils as scrub_utils
import analysis_engine.get_data_from_redis_key as redis_get
import analysis_engine.td.consts as td_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def extract_option_calls_dataset(
        work_dict,
        scrub_mode='sort-by-date'):
    """extract_option_calls_dataset

    Extract the TD options calls for a ticker and
    return it as a ``pandas.Dataframe``

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = '{}'.format(work_dict.get('label', 'extract'))
    ds_id = work_dict.get('ticker')
    df_type = td_consts.DATAFEED_TD_CALLS
    df_str = td_consts.get_datafeed_str_td(df_type=df_type)
    redis_key = work_dict.get(
        'redis_key',
        work_dict.get('tdcalls', 'missing-redis-key'))
    s3_key = work_dict.get(
        's3_key',
        work_dict.get('tdcalls', 'missing-s3-key'))
    redis_host = work_dict.get(
        'redis_host',
        None)
    redis_port = work_dict.get(
        'redis_port',
        None)
    redis_db = work_dict.get(
        'redis_db',
        ae_consts.REDIS_DB)
    verbose = work_dict.get(
        'verbose_td',
        False)

    if verbose:
        log.info(
            '{} - {} - start - redis_key={} s3_key={}'.format(
                label,
                df_str,
                redis_key,
                s3_key))

    if not redis_host and not redis_port:
        redis_host = ae_consts.REDIS_ADDRESS.split(':')[0]
        redis_port = ae_consts.REDIS_ADDRESS.split(':')[1]

    exp_date_str = None
    calls_df = None
    status = ae_consts.NOT_RUN
    try:
        redis_rec = redis_get.get_data_from_redis_key(
            label=label,
            host=redis_host,
            port=redis_port,
            db=redis_db,
            password=work_dict.get('password', None),
            key=redis_key,
            decompress_df=True)

        status = redis_rec['status']
        if verbose:
            log.info(
                '{} - {} redis get data key={} status={}'.format(
                    label,
                    df_str,
                    redis_key,
                    ae_consts.get_status(status=status)))

        if status == ae_consts.SUCCESS:
            calls_json = None
            if 'calls' in redis_rec['rec']['data']:
                calls_json = redis_rec['rec']['data']['calls']
            else:
                calls_json = redis_rec['rec']['data']
            if verbose:
                log.info(
                    '{} - {} redis convert calls to df'.format(
                        label,
                        df_str))
            exp_date_str = None
            try:
                calls_df = pd.read_json(
                    calls_json,
                    orient='records')
                if len(calls_df.index) == 0:
                    return ae_consts.SUCCESS, None
                if 'date' not in calls_df:
                    log.debug(
                        'failed to find date column in TD calls '
                        'df={}'.format(
                            calls_df,
                            len(calls_df.index)))
                    return ae_consts.SUCCESS, None
                calls_df.sort_values(
                        by=[
                            'date',
                            'strike'
                        ])
                """
                for i, r in calls_df.iterrows():
                    print(r['date'])
                convert_epochs = [
                    'ask_date',
                    'bid_date',
                    'trade_date'
                ]
                for c in convert_epochs:
                    if c in calls_df:
                        calls_df[c] = pd.DatetimeIndex(pd.to_datetime(
                            calls_df[c],
                            format=ae_consts.COMMON_TICK_DATE_FORMAT
                        )).tz_localize(
                            'UTC').tz_convert(
                                'US/Eastern')
                # dates converted
                """
                exp_date_str = (
                    calls_df['exp_date'].iloc[-1])

                calls_df['date'] = calls_df['date'].dt.strftime(
                    ae_consts.COMMON_TICK_DATE_FORMAT)

            except Exception as f:
                log.error(
                    '{} - {} redis_key={} '
                    'no calls df found or ex={}'.format(
                        label,
                        df_str,
                        redis_key,
                        f))
                return ae_consts.EMPTY, None
            # end of try/ex to convert to df
            if verbose:
                log.info(
                    '{} - {} redis_key={} calls={} exp_date={}'.format(
                        label,
                        df_str,
                        redis_key,
                        len(calls_df.index),
                        exp_date_str))
        else:
            if verbose:
                log.info(
                    '{} - {} did not find valid redis option calls '
                    'in redis_key={} status={}'.format(
                        label,
                        df_str,
                        redis_key,
                        ae_consts.get_status(status=status)))

    except Exception as e:
        log.debug(
            '{} - {} - ds_id={} failed getting option calls from '
            'redis={}:{}@{} key={} ex={}'.format(
                label,
                df_str,
                ds_id,
                redis_host,
                redis_port,
                redis_db,
                redis_key,
                e))
        return ae_consts.ERR, None
    # end of try/ex extract from redis

    if verbose:
        log.info(
            '{} - {} ds_id={} extract scrub={}'.format(
                label,
                df_str,
                ds_id,
                scrub_mode))

    scrubbed_df = scrub_utils.extract_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=df_type,
        msg_format='df={} date_str={}',
        ds_id=ds_id,
        df=calls_df)

    status = ae_consts.SUCCESS

    return status, scrubbed_df
# end of extract_option_calls_dataset


def extract_option_puts_dataset(
        work_dict,
        scrub_mode='sort-by-date'):
    """extract_option_puts_dataset

    Extract the TD options puts for a ticker and
    return it as a ``pandas.Dataframe``

    :param work_dict: dictionary of args
    :param scrub_mode: type of scrubbing handler to run
    """
    label = '{}'.format(work_dict.get('label', 'extract'))
    ds_id = work_dict.get('ticker')
    df_type = td_consts.DATAFEED_TD_PUTS
    df_str = td_consts.get_datafeed_str_td(df_type=df_type)
    redis_key = work_dict.get(
        'redis_key',
        work_dict.get('tdputs', 'missing-redis-key'))
    s3_key = work_dict.get(
        's3_key',
        work_dict.get('tdputs', 'missing-s3-key'))
    redis_host = work_dict.get(
        'redis_host',
        None)
    redis_port = work_dict.get(
        'redis_port',
        None)
    redis_db = work_dict.get(
        'redis_db',
        ae_consts.REDIS_DB)
    verbose = work_dict.get(
        'verbose_td',
        False)

    if verbose:
        log.info(
            '{} - {} - start - redis_key={} s3_key={}'.format(
                label,
                df_str,
                redis_key,
                s3_key))

    if not redis_host and not redis_port:
        redis_host = ae_consts.REDIS_ADDRESS.split(':')[0]
        redis_port = ae_consts.REDIS_ADDRESS.split(':')[1]

    exp_date_str = None
    puts_df = None
    status = ae_consts.NOT_RUN
    try:
        redis_rec = redis_get.get_data_from_redis_key(
            label=label,
            host=redis_host,
            port=redis_port,
            db=redis_db,
            password=work_dict.get('password', None),
            key=redis_key,
            decompress_df=True)

        status = redis_rec['status']
        if verbose:
            log.info(
                '{} - {} redis get data key={} status={}'.format(
                    label,
                    df_str,
                    redis_key,
                    ae_consts.get_status(status=status)))

        if status == ae_consts.SUCCESS:
            puts_json = None
            if 'puts' in redis_rec['rec']['data']:
                puts_json = redis_rec['rec']['data']['puts']
            else:
                puts_json = redis_rec['rec']['data']
            if verbose:
                log.info(
                    '{} - {} redis convert puts to df'.format(
                        label,
                        df_str))
            try:
                puts_df = pd.read_json(
                    puts_json,
                    orient='records')
                if len(puts_df.index) == 0:
                    return ae_consts.SUCCESS, None
                if 'date' not in puts_df:
                    log.debug(
                        'failed to find date column in TD puts '
                        'df={} len={}'.format(
                            puts_df,
                            len(puts_df.index)))
                    return ae_consts.SUCCESS, None
                puts_df.sort_values(
                        by=[
                            'date',
                            'strike'
                        ])
                """
                for i, r in calls_df.iterrows():
                    print(r['date'])
                convert_epochs = [
                    'ask_date',
                    'bid_date',
                    'trade_date'
                ]
                for c in convert_epochs:
                    if c in puts_df:
                        puts_df[c] = pd.DatetimeIndex(pd.to_datetime(
                            puts_df[c],
                            format=ae_consts.COMMON_TICK_DATE_FORMAT
                        )).tz_localize(
                            'UTC').tz_convert(
                                'US/Eastern')
                # dates converted
                """
                exp_date_str = (
                    puts_df['exp_date'].iloc[-1])

                puts_df['date'] = puts_df['date'].dt.strftime(
                    ae_consts.COMMON_TICK_DATE_FORMAT)

            except Exception as f:
                log.debug(
                    '{} - {} redis_key={} '
                    'no puts df found'.format(
                        label,
                        df_str,
                        redis_key))
                return ae_consts.EMPTY, None
            # end of try/ex to convert to df
            if verbose:
                log.info(
                    '{} - {} redis_key={} puts={} exp_date={}'.format(
                        label,
                        df_str,
                        redis_key,
                        len(puts_df.index),
                        exp_date_str))
        else:
            if verbose:
                log.info(
                    '{} - {} did not find valid redis option puts '
                    'in redis_key={} status={}'.format(
                        label,
                        df_str,
                        redis_key,
                        ae_consts.get_status(status=status)))

    except Exception as e:
        log.debug(
            '{} - {} - ds_id={} failed getting option puts from '
            'redis={}:{}@{} key={} ex={}'.format(
                label,
                df_str,
                ds_id,
                redis_host,
                redis_port,
                redis_db,
                redis_key,
                e))
        return ae_consts.ERR, None
    # end of try/ex extract from redis

    if verbose:
        log.info(
            '{} - {} ds_id={} extract scrub={}'.format(
                label,
                df_str,
                ds_id,
                scrub_mode))

    scrubbed_df = scrub_utils.extract_scrub_dataset(
        label=label,
        scrub_mode=scrub_mode,
        datafeed_type=df_type,
        msg_format='df={} date_str={}',
        ds_id=ds_id,
        df=puts_df)

    status = ae_consts.SUCCESS

    return status, scrubbed_df
# end of extract_option_puts_dataset



"""
Mock redis objects
"""

import analysis_engine.consts as ae_consts
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


class MockRedisFailToConnect:
    """MockRedisFailToConnect"""

    def __init__(
            self,
            host=None,
            port=None,
            password=None,
            db=None):
        """__init__

        build a mock redis client that will raise an exception
        to test failures during connection

        :param host: hostname
        :param port: port
        :param password: password
        :param db: database number
        """
        raise Exception(
            'test MockRedisFailToConnect')
    # end of __init__

# end of MockRedisFailToConnect


class MockRedis:
    """MockRedis"""

    def __init__(
            self,
            host=None,
            port=None,
            password=None,
            db=None):
        """__init__

        build mock redis client

        :param host: hostname
        :param port: port
        :param password: password
        :param db: database number
        """
        self.host = host
        self.port = port
        self.password = password
        self.db = db
        self.cache_dict = {}  # cache dictionary replicating redis
        self.keys = []        # cache redis keys
    # end of __init__

    def set(
            self,
            name=None,
            value=None,
            ex=None,
            px=None,
            nx=False,
            xx=False):
        """set

        mock redis set

        :param name: cache key name
        :param value: value to cache
        :param ex: expire time
        :param px: redis values
        :param nx: redis values
        :param xx: redis values
        """

        log.info(
            'mock - MockRedis.set('
            'name={}, '
            'value={}, '
            'ex={}, '
            'px={}, '
            'nx={}, '
            'xx={})'.format(
                name,
                value,
                ex,
                px,
                nx,
                xx))
        self.cache_dict[name] = value
        self.keys.append(name)
    # end of set

    def get(
            self,
            name=None):
        """get

        mock redis get

        :param name: name of the key to check
        """
        if not name:
            err = (
                'mock - MockRedis.get('
                'name={}'
                ') - missing a name'.format(
                    name))
            log.error(err)
            raise Exception(err)
        value_in_dict = self.cache_dict.get(
            name,
            None)
        if not value_in_dict:
            value_in_env = ae_consts.ev(
                name,
                None)
            log.info(
                'mock - MockRedis.get('
                'name={}) '
                'env={}'.format(
                    name,
                    value_in_env))
            if value_in_env:
                return value_in_env.encode('utf-8')
            else:
                return None
        else:
            log.info(
                'mock - MockRedis.get('
                'name={}) '
                'cached_value={}'.format(
                    name,
                    value_in_dict))
            return value_in_dict
        # end of get data from dict vs in the env
    # end of get

# end of MockRedis



"""
Example Minute Algorithm for showing how
to run an algorithm on intraday minute timeseries datasets

**What does the base class provide?**

Algorithms automatically provide the following
member variables to any custom algorithm that derives
the ``analysis_engine.algo.BaseAlgo.process`` method.

By deriving the ``process()`` member method using an inherited
class, you can quickly build algorithms that
determine **buy** and **sell** conditions from
any of the automatically extracted
datasets from the redis pipeline:

- ``self.df_daily``
- ``self.df_minute``
- ``self.df_calls``
- ``self.df_puts``
- ``self.df_quote``
- ``self.df_pricing``
- ``self.df_stats``
- ``self.df_peers``
- ``self.df_iex_news``
- ``self.df_financials``
- ``self.df_earnings``
- ``self.df_dividends``
- ``self.df_company``
- ``self.df_yahoo_news``

**Recent Pricing Information**

- ``self.latest_close``
- ``self.latest_high``
- ``self.latest_open``
- ``self.latest_low``
- ``self.latest_volume``
- ``self.ask``
- ``self.bid``

**Latest Backtest Date and Intraday Minute**

- ``self.latest_min``
- ``self.backtest_date``

.. note:: **self.latest_min** - Latest minute row in ``self.df_minute``

.. note:: **self.backtest_date** - Latest dataset date which is considered the
    backtest date for historical testing with the data pipeline
    structure (it's the ``date`` key in the dataset node root level)

**Balance Information**

- ``self.balance``
- ``self.prev_bal``

.. note:: If a key is not in the dataset, the
    algorithms's member variable will be an empty
    pandas DataFrame created with: ``pd.DataFrame([])``
    except ``self.pricing`` which is just a dictionary.
    Please ensure the engine successfully fetched
    and cached the dataset in redis using a tool like
    ``redis-cli`` and a query of ``keys *`` or
    ``keys <TICKER>_*`` on large deployments.

**Supported environment variables**

::

    # to show debug, trace logging please export ``SHARED_LOG_CFG``
    # to a debug logger json file. To turn on debugging for this
    # library, you can export this variable to the repo's
    # included file with the command:
    export SHARED_LOG_CFG=/opt/sa/analysis_engine/log/debug-logging.json
"""

import analysis_engine.algo as base_algo
import analysis_engine.utils as ae_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


class ExampleMinuteAlgo(base_algo.BaseAlgo):
    """ExampleMinuteAlgo"""

    def __init__(
            self,
            **kwargs):
        """__init__

        Custom ticker algo for analyzing intraday
        minute datasets

        Please refer to the `analysis_engine.algo.Ba
        seAlgo source code for the latest supported parameters <ht
        tps://github.com/AlgoTraders/stock-analysis-engine
        /blob/master/
        analysis_engine/algo.py>`__

        :param kwargs: keyword arguments
        """
        log.info('base - {}'.format(kwargs.get('name', 'no-name-set')))
        super().__init__(**kwargs)
        log.info('ready - {}'.format(self.name))
    # end of __init__

    def process(
            self,
            algo_id,
            ticker,
            dataset):
        """process

        Derive custom algorithm buy and sell conditions
        before placing orders. Just implement your own
        ``process`` method.

        :param algo_id: string - algo identifier label for debugging datasets
            during specific dates
        :param ticker: string - ticker
        :param dataset: a dictionary of identifiers (for debugging) and
            multiple pandas ``pd.DataFrame`` objects. Dictionary where keys
            represent a label from one of the data sources (``IEX``,
            ``Yahoo``, ``FinViz`` or other). Here is the supported
            dataset structure for the process method:

            .. note:: There are no required keys for ``data``, the list
                below is not hard-enforced by default. This is just
                a reference for what is available with the v1 engine.

            ::

                dataset = {
                    'id': <string TICKER_DATE - redis cache key>,
                    'date': <string DATE>,
                    'data': {
                        'daily': pd.DataFrame([]),
                        'minute': pd.DataFrame([]),
                        'quote': pd.DataFrame([]),
                        'stats': pd.DataFrame([]),
                        'peers': pd.DataFrame([]),
                        'news1': pd.DataFrame([]),
                        'financials': pd.DataFrame([]),
                        'earnings': pd.DataFrame([]),
                        'dividends': pd.DataFrame([]),
                        'calls': pd.DataFrame([]),
                        'puts': pd.DataFrame([]),
                        'pricing': dictionary,
                        'news': pd.DataFrame([])
                    }
                }

            example:

            ::

                dataset = {
                    'id': 'SPY_2018-11-02
                    'date': '2018-11-02',
                    'data': {
                        'daily': pd.DataFrame,
                        'minute': pd.DataFrame,
                        'calls': pd.DataFrame,
                        'puts': pd.DataFrame,
                        'news': pd.DataFrame
                    }
                }
        """
        num_minute_rows = len(self.df_minute.index)
        label = (
            'process - {} - ticker={}'.format(
                self.name,
                ticker))
        log.info(
            '{} - start - '
            'date={} minute={} close={} '
            'high={} low={} open={} volume={} '
            'intraday rows={}'.format(
                label,
                self.backtest_date,
                self.latest_min,
                self.latest_close,
                self.latest_high,
                self.latest_low,
                self.latest_open,
                self.latest_volume,
                num_minute_rows))

        # walk through today's intraday records:
        # Provided by IEX
        num_done = 1
        total_records = num_minute_rows
        for idx, row in self.df_minute.iterrows():
            percent_label = self.build_progress_label(
                progress=num_done,
                total=total_records)
            if 'date' in row:
                # want to debug rows on stdout?
                # print(row)
                log.info(
                    '{} - ANALYZE - '
                    'date={} at {} '
                    'close={} high={} low={} open={} volume={} '
                    '- {}'.format(
                        label,
                        self.backtest_date,
                        row['date'].strftime('%H:%M:%S'),
                        row['close'],
                        row['high'],
                        row['low'],
                        row['open'],
                        row['volume'],
                        percent_label))
                num_done += 1
        # end for all rows in the dataset

        log.info(
            '{} - done - '
            'date={} minute={} '
            'balance={} previous_balance={} '
            'shares={} buys={} sells={} '
            'close={} high={} low={} open={} volume={} '
            'intraday rows={}'.format(
                label,
                self.backtest_date,
                self.latest_min,
                self.balance,
                self.prev_bal,
                self.num_owned,
                self.num_buys,
                self.num_sells,
                self.latest_close,
                self.latest_high,
                self.latest_low,
                self.latest_open,
                self.latest_volume,
                num_minute_rows))

    # end of process

    def get_result(
            self):
        """get_result"""

        log.info('building results')
        finished_date = ae_utils.utc_now_str()
        self.result = {
            'name': self.name,
            'created': self.created_date,
            'updated': finished_date,
            'open_positions': self.positions,
            'buys': self.get_buys(),
            'sells': self.get_sells(),
            'num_processed': len(self.order_history),
            'history': self.order_history,
            'balance': self.balance,
            'commission': self.commission
        }

        return self.result
    # end of get_result

# end of ExampleMinuteAlgo


def get_algo(
        **kwargs):
    """get_algo

    Make sure to define the ``get_algo`` for your custom
    algorithms to work as a backup with the ``sa.py`` tool...
    Not anticipating issues, but if we do with importlib
    this is the backup plan.

    Please file an issue if you see something weird and would like
    some help:
    https://github.com/AlgoTraders/stock-analysis-engine/issues

    :param kwargs: dictionary of keyword arguments
    """
    log.info('getting algo')
    return ExampleMinuteAlgo(**kwargs)
# end of get_algo



"""
Helpful wrapper taken from:
https://www.peterbe.com/plog/best-practice-with-retries-with-requests
"""

import requests
import requests.adapters as adapters
import requests.packages.urllib3.util.retry as requests_retry


def url_helper(
        sess=None,
        retries=10,
        backoff_factor=0.3,
        status_forcelist=(500, 502, 504)):
    """url_helper

    :param sess: ``requests.Session``
        object like

        .. code-block:: python

            s = requests.Session()
            s.auth = ('user', 'pass')
            s.headers.update({'x-test': 'true'})

            response = url_helper(sesssion=s).get(
                'https://www.peterbe.com'
            )

    :param retries: number of retries
        default is ``3``
    :param backoff_factor: seconds per attempt
        default is ``0.3``
    :param status_forcelist: optional tuple list
        of retry error HTTP status codes
        default is ``500, 502, 504``
    """
    session = sess or requests.Session()
    retry = requests_retry.Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = adapters.HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session
# end of url_helper



"""
Mocking data fetch api calls
"""

import datetime


def mock_daily(
        url,
        token=None,
        version=None):
    """mock_daily

    mock minute history for a chart

    :param url: IEX resource url
    :param token: optional - string token for your user's
        account
    :param version: optional - version string
    """

    val = {
        'url': url,
        'version': version,
        'symbol': url.split('/')[2],
        'timeframe': '3m',
        'date': datetime.datetime.now().strftime('%Y-%m-%d'),
        'testcase': 'mock-daily'
    }
    return [val]
# end of mock_daily


def mock_minute(
        url,
        token=None,
        version=None):
    """mock_minute

    mock minute history for a chart

    :param url: IEX resource url
    :param token: optional - string token for your user's
        account
    :param version: optional - version string
    """

    now = datetime.datetime.now()
    val = {
        'url': url,
        'version': version,
        'symbol': url.split('/')[2],
        'timeframe': '1d',
        'date': now.strftime('%Y-%m-%d'),
        'minute': now.strftime('%H:%M'),
        'testcase': 'mock-minute'

    }
    return [val]
# end of mock_minute


def mock_quote(
        url,
        token=None,
        version=None):
    """mock_quote

    mock quote

    :param url: IEX resource url
    :param token: optional - string token for your user's
        account
    :param version: optional - version string
    """

    val = {
        'url': url,
        'version': version,
        'symbol': url.split('/')[2],
        'testcase': 'mock-quote'
    }
    return val
# end of mock_quote


def mock_stats(
        url,
        token=None,
        version=None):
    """mock_stats

    mock stats

    :param url: IEX resource url
    :param token: optional - string token for your user's
        account
    :param version: optional - version string
    """

    val = {
        'url': url,
        'version': version,
        'symbol': url.split('/')[2],
        'testcase': 'mock-stats'
    }
    return val
# end of mock_stats


def mock_peers(
        url,
        token=None,
        version=None):
    """mock_peers

    mock peers

    :param url: IEX resource url
    :param token: optional - string token for your user's
        account
    :param version: optional - version string
    """

    val = {
        'url': url,
        'version': version,
        'symbol': url.split('/')[2],
        'testcase': 'mock-peers'
    }
    return [val]
# end of mock_peers


def mock_news(
        url,
        token=None,
        version=None):
    """mock_news

    mock news

    :param url: IEX resource url
    :param token: optional - string token for your user's
        account
    :param version: optional - version string
    """
    now = datetime.datetime.now()
    epoch = datetime.datetime.utcfromtimestamp(0)
    now_ms = (now - epoch).total_seconds() * 1000.0

    val = {
        'url': url,
        'version': version,
        'datetime': now_ms,
        'symbol': url.split('/')[2],
        'count': 5,
        'testcase': 'mock-news'
    }
    return [val]
# end of mock_news


def mock_financials(
        url,
        token=None,
        version=None):
    """mock_financials

    mock financials

    :param url: IEX resource url
    :param token: optional - string token for your user's
        account
    :param version: optional - version string
    """

    val = {
        'url': url,
        'version': version,
        'symbol': url.split('/')[2],
        'testcase': 'mock-financials'
    }
    return {
        'financials': [val]
    }
# end of mock_financials


def mock_earnings(
        url,
        token=None,
        version=None):
    """mock_earnings

    mock earnings

    :param url: IEX resource url
    :param token: optional - string token for your user's
        account
    :param version: optional - version string
    """

    val = {
        'url': url,
        'version': version,
        'symbol': url.split('/')[2],
        'testcase': 'mock-earnings'
    }
    return {
        'earnings': [val]
    }
# end of mock_earnings


def mock_dividends(
        url,
        token=None,
        version=None):
    """mock_dividends

    mock dividends

    :param url: IEX resource url
    :param token: optional - string token for your user's
        account
    :param version: optional - version string
    """

    val = {
        'url': url,
        'version': version,
        'symbol': url.split('/')[2],
        'testcase': 'mock-dividends'
    }
    return [val]
# end of mock_dividends


def mock_company(
        url,
        token=None,
        version=None):
    """mock_company

    mock company

    :param url: IEX resource url
    :param token: optional - string token for your user's
        account
    :param version: optional - version string
    """

    val = {
        'url': url,
        'version': version,
        'symbol': url.split('/')[2],
        'testcase': 'mock-company'
    }
    return val
# end of mock_company



"""
**Prepare Pricing Dataset**

Prepare dataset for analysis. This task collapses
nested json dictionaries into a
csv file with a header row and stores the output
file in s3 and redis automatically.

- if key not in redis, load the key by the same name from s3
- prepare dataset from redis key
- the dataset will be stored as a dictionary with a pandas dataframe

**Sample work_dict request for this method**

`analysis_engine.api_requests.build_prepare_dataset_request <https://
github.com/AlgoTraders/stock-analysis-engine/blob/master/
analysis_engine/api_requests.py#L300>`__

::

    work_request = {
        'ticker': ticker,
        'ticker_id': ticker_id,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        'prepared_s3_key': s3_prepared_key,
        'prepared_s3_bucket': s3_prepared_bucket_name,
        'prepared_redis_key': redis_prepared_key,
        'ignore_columns': ignore_columns,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled
    }

.. tip:: This task uses the `analysis_engine.work_tasks.
    custom_task.CustomTask class <https://github.com/A
    lgoTraders/stock-analysis-engine/blob/master/anal
    ysis_engine/work_tasks/custom_task.py>`__ for
    task event handling.

**Supported Environment Variables**

::

    export DEBUG_PREPARE=1
    export DEBUG_RESULTS=1

"""

import datetime
import redis
import celery.task as celery_task
import analysis_engine.consts as ae_consts
import analysis_engine.build_result as build_result
import analysis_engine.api_requests as api_requests
import analysis_engine.get_data_from_redis_key as redis_get
import analysis_engine.get_task_results as get_task_results
import analysis_engine.work_tasks.custom_task as custom_task
import analysis_engine.work_tasks.publish_from_s3_to_redis as s3_to_redis
import analysis_engine.dict_to_csv as dict_to_csv
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


@celery_task(
    bind=True,
    base=custom_task.CustomTask,
    queue='prepare_pricing_dataset')
def prepare_pricing_dataset(
        self,
        work_dict):
    """prepare_pricing_dataset

    Prepare dataset for analysis. Supports loading dataset from
    s3 if not found in redis. Outputs prepared artifact as a csv
    to s3 and redis.

    :param work_dict: dictionary for key/values
    """

    label = 'prepare'

    log.info(
        'task - {} - start '
        'work_dict={}'.format(
            label,
            work_dict))

    initial_data = None

    ticker = ae_consts.TICKER
    ticker_id = ae_consts.TICKER_ID
    rec = {
        'ticker': None,
        'ticker_id': None,
        's3_enabled': True,
        'redis_enabled': True,
        's3_bucket': None,
        's3_key': None,
        'redis_key': None,
        'prepared_s3_key': None,
        'prepared_s3_bucket': None,
        'prepared_redis_key': None,
        'prepared_data': None,
        'prepared_size': None,
        'initial_data': None,
        'initial_size': None,
        'ignore_columns': None,
        'updated': None
    }
    res = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec=rec)

    try:
        ticker = work_dict.get(
            'ticker',
            ae_consts.TICKER)
        ticker_id = int(work_dict.get(
            'ticker_id',
            ae_consts.TICKER_ID))

        if not ticker:
            res = build_result.build_result(
                status=ae_consts.ERR,
                err='missing ticker',
                rec=rec)
            return res

        label = work_dict.get(
            'label',
            label)
        s3_key = work_dict.get(
            's3_key',
            None)
        s3_bucket_name = work_dict.get(
            's3_bucket',
            'pricing')
        s3_access_key = work_dict.get(
            's3_access_key',
            ae_consts.S3_ACCESS_KEY)
        s3_secret_key = work_dict.get(
            's3_secret_key',
            ae_consts.S3_SECRET_KEY)
        s3_region_name = work_dict.get(
            's3_region_name',
            ae_consts.S3_REGION_NAME)
        s3_address = work_dict.get(
            's3_address',
            ae_consts.S3_ADDRESS)
        s3_secure = work_dict.get(
            's3_secure',
            ae_consts.S3_SECURE) == '1'
        redis_address = work_dict.get(
            'redis_address',
            ae_consts.REDIS_ADDRESS)
        redis_key = work_dict.get(
            'redis_key',
            ae_consts.REDIS_KEY)
        redis_password = work_dict.get(
            'redis_password',
            ae_consts.REDIS_PASSWORD)
        redis_db = work_dict.get(
            'redis_db',
            None)
        if not redis_db:
            redis_db = ae_consts.REDIS_DB
        redis_expire = None
        if 'redis_expire' in work_dict:
            redis_expire = work_dict.get(
                'redis_expire',
                ae_consts.REDIS_EXPIRE)
        updated = work_dict.get(
            'updated',
            datetime.datetime.utcnow().strftime(
                '%Y_%m_%d_%H_%M_%S'))
        prepared_s3_key = work_dict.get(
            'prepared_s3_key',
            '{}_{}.csv'.format(
                ticker,
                updated))
        prepared_s3_bucket = work_dict.get(
            'prepared_s3_bucket',
            'prepared')
        prepared_redis_key = work_dict.get(
            'prepared_redis_key',
            'prepared')
        ignore_columns = work_dict.get(
            'ignore_columns',
            None)
        log.info(
            '{} redis enabled address={}@{} '
            'key={} prepare_s3={}:{} prepare_redis={} '
            'ignore_columns={}'.format(
                label,
                redis_address,
                redis_db,
                redis_key,
                prepared_s3_bucket,
                prepared_s3_key,
                prepared_redis_key,
                ignore_columns))
        redis_host = redis_address.split(':')[0]
        redis_port = redis_address.split(':')[1]

        enable_s3 = True
        enable_redis_publish = True

        rec['ticker'] = ticker
        rec['ticker_id'] = ticker_id
        rec['s3_bucket'] = s3_bucket_name
        rec['s3_key'] = s3_key
        rec['redis_key'] = redis_key
        rec['prepared_s3_key'] = prepared_s3_key
        rec['prepared_s3_bucket'] = prepared_s3_bucket
        rec['prepared_redis_key'] = prepared_redis_key
        rec['updated'] = updated
        rec['s3_enabled'] = enable_s3
        rec['redis_enabled'] = enable_redis_publish

        try:
            log.info(
                '{} connecting redis={}:{} '
                'db={} key={} '
                'updated={} expire={}'.format(
                    label,
                    redis_host,
                    redis_port,
                    redis_db,
                    redis_key,
                    updated,
                    redis_expire))
            rc = redis.Redis(
                host=redis_host,
                port=redis_port,
                password=redis_password,
                db=redis_db)
        except Exception as e:
            err = (
                '{} failed - redis connection to address={}@{} '
                'key={} ex={}'.format(
                    label,
                    redis_address,
                    redis_key,
                    redis_db,
                    e))
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res
        # end of try/ex for connecting to redis

        initial_data_res = redis_get.get_data_from_redis_key(
            label=label,
            client=rc,
            key=redis_key)

        log.info(
            '{} get redis key={} status={} err={}'.format(
                label,
                redis_key,
                ae_consts.get_status(initial_data_res['status']),
                initial_data_res['err']))

        initial_data = initial_data_res['rec'].get(
            'data',
            None)

        if enable_s3 and not initial_data:

            log.info(
                '{} failed to find redis_key={} trying s3 '
                'from s3_key={} s3_bucket={} s3_address={}'.format(
                    label,
                    redis_key,
                    s3_key,
                    s3_bucket_name,
                    s3_address))

            get_from_s3_req = \
                api_requests.build_publish_from_s3_to_redis_request()

            get_from_s3_req['s3_enabled'] = enable_s3
            get_from_s3_req['s3_access_key'] = s3_access_key
            get_from_s3_req['s3_secret_key'] = s3_secret_key
            get_from_s3_req['s3_region_name'] = s3_region_name
            get_from_s3_req['s3_address'] = s3_address
            get_from_s3_req['s3_secure'] = s3_secure
            get_from_s3_req['s3_key'] = s3_key
            get_from_s3_req['s3_bucket'] = s3_bucket_name
            get_from_s3_req['redis_key'] = redis_key
            get_from_s3_req['label'] = (
                '{}-run_publish_from_s3_to_redis'.format(
                    label))

            log.info(
                '{} load from s3={} to '
                'redis={}'.format(
                    label,
                    s3_key,
                    redis_key))

            try:
                # run in synchronous mode:
                get_from_s3_req['celery_disabled'] = True
                task_res = s3_to_redis.run_publish_from_s3_to_redis(
                    get_from_s3_req)
                if task_res.get(
                        'status',
                        ae_consts.ERR) == ae_consts.SUCCESS:
                    log.info(
                        '{} loaded s3={}:{} '
                        'to redis={} retrying'.format(
                            label,
                            s3_bucket_name,
                            s3_key,
                            redis_key))
                    initial_data_res = redis_get.get_data_from_redis_key(
                        label=label,
                        client=rc,
                        key=redis_key)

                    log.info(
                        '{} get redis try=2 key={} status={} err={}'.format(
                            label,
                            redis_key,
                            ae_consts.get_status(initial_data_res['status']),
                            initial_data_res['err']))

                    initial_data = initial_data_res['rec'].get(
                        'data',
                        None)
                else:
                    err = (
                        '{} ERR failed loading from bucket={} '
                        's3_key={} to redis_key={} with res={}'.format(
                            label,
                            s3_bucket_name,
                            s3_key,
                            redis_key,
                            task_res))
                    log.error(err)
                    res = build_result.build_result(
                        status=ae_consts.ERR,
                        err=err,
                        rec=rec)
                    return res
            except Exception as e:
                err = (
                    '{} extract from s3 and publish to redis failed loading '
                    'data from bucket={} in '
                    's3_key={} with publish to redis_key={} '
                    'with ex={}'.format(
                        label,
                        s3_bucket_name,
                        s3_key,
                        redis_key,
                        e))
                log.error(err)
                res = build_result.build_result(
                    status=ae_consts.ERR,
                    err=err,
                    rec=rec)
                return res
            # end of try/ex for publishing from s3->redis
        # end of if enable_s3

        if not initial_data:
            err = (
                '{} did not find any data to prepare in redis_key={} or '
                's3_key={} in bucket={}'.format(
                    label,
                    redis_key,
                    s3_key,
                    s3_bucket_name))
            log.error(err)
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res

        initial_data_num_chars = len(str(initial_data))
        initial_size_value = None
        initial_size_str = None
        if initial_data_num_chars < ae_consts.PREPARE_DATA_MIN_SIZE:
            err = (
                '{} not enough data={} in redis_key={} or '
                's3_key={} in bucket={}'.format(
                    label,
                    initial_data_num_chars,
                    redis_key,
                    s3_key,
                    s3_bucket_name))
            log.error(err)
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res
        else:
            initial_size_value = initial_data_num_chars / 1024000
            initial_size_str = ae_consts.to_f(initial_size_value)
            if ae_consts.ev('DEBUG_PREPARE', '0') == '1':
                log.info(
                    '{} initial - redis_key={} data={}'.format(
                        label,
                        redis_key,
                        str(initial_data)))
            else:
                log.info(
                    '{} initial - redis_key={} data size={} MB'.format(
                        label,
                        redis_key,
                        initial_size_str))
        # end of trying to get initial_data

        rec['initial_data'] = initial_data
        rec['initial_size'] = initial_data_num_chars

        prepare_data = None

        try:
            if ae_consts.ev('DEBUG_PREPARE', '0') == '1':
                log.info(
                    '{} data={} - flatten - {} MB from '
                    'redis_key={}'.format(
                        label,
                        ae_consts.ppj(initial_data),
                        initial_size_str,
                        redis_key))
            else:
                log.info(
                    '{} flatten - {} MB from '
                    'redis_key={}'.format(
                        label,
                        initial_size_str,
                        redis_key))
            prepare_data = dict_to_csv.flatten_dict(
                data=initial_data)
        except Exception as e:
            prepare_data = None
            err = (
                '{} flatten - convert to csv failed with ex={} '
                'redis_key={}'.format(
                    label,
                    e,
                    redis_key))
            log.error(err)
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res
        # end of try/ex

        if not prepare_data:
            err = (
                '{} flatten - did not return any data from redis_key={} '
                'or s3_key={} in bucket={}'.format(
                    label,
                    redis_key,
                    s3_key,
                    s3_bucket_name))
            log.error(err)
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res
        # end of prepare_data

        prepare_data_num_chars = len(str(prepare_data))
        prepare_size_value = None

        if prepare_data_num_chars < ae_consts.PREPARE_DATA_MIN_SIZE:
            err = (
                '{} prepare - there is not enough data={} in redis_key={}'
                ''.format(
                    label,
                    prepare_data_num_chars,
                    redis_key))
            log.error(err)
            res = build_result.build_result(
                status=ae_consts.ERR,
                err=err,
                rec=rec)
            return res
        else:
            prepare_size_value = prepare_data_num_chars / 1024000
            prepare_size_str = ae_consts.to_f(prepare_size_value)
            if ae_consts.ev('DEBUG_PREPARE', '0') == '1':
                log.info(
                    '{} data={} - prepare - redis_key={}'.format(
                        label,
                        redis_key,
                        ae_consts.ppj(prepare_data)))
            else:
                log.info(
                    '{} prepare - redis_key={} data size={} MB'.format(
                        label,
                        redis_key,
                        prepare_size_str))
        # end of trying to the size of the prepared data

        rec['prepared_data'] = prepare_data
        rec['prepared_size'] = prepare_data_num_chars

        res = build_result.build_result(
            status=ae_consts.SUCCESS,
            err=None,
            rec=rec)

        rc = None

    except Exception as e:
        res = build_result.build_result(
            status=ae_consts.ERR,
            err=(
                'failed - prepare_pricing_dataset '
                'dict={} with ex={}').format(
                    work_dict,
                    e),
            rec=rec)
        log.error(
            '{} - {}'.format(
                label,
                res['err']))
    # end of try/ex

    log.info(
        'task - prepare_pricing_dataset done - '
        '{} - status={}'.format(
            label,
            ae_consts.get_status(res['status'])))

    return get_task_results.get_task_results(
        work_dict=work_dict,
        result=res)
# end of prepare_pricing_dataset


def run_prepare_pricing_dataset(
        work_dict):
    """run_prepare_pricing_dataset

    Celery wrapper for running without celery

    :param work_dict: task data
    """

    label = work_dict.get(
        'label',
        '')

    log.info(
        'run_prepare_pricing_dataset - {} - start'.format(
            label))

    response = build_result.build_result(
        status=ae_consts.NOT_RUN,
        err=None,
        rec={})
    task_res = {}

    # allow running without celery
    if ae_consts.is_celery_disabled(
            work_dict=work_dict):
        work_dict['celery_disabled'] = True
        task_res = prepare_pricing_dataset(
            work_dict)
        if task_res:
            response = task_res.get(
                'result',
                task_res)
            if ae_consts.ev('DEBUG_RESULTS', '0') == '1':
                response_details = response
                try:
                    response_details = ae_consts.ppj(response)
                except Exception:
                    response_details = response
                log.info(
                    '{} task result={}'.format(
                        label,
                        response_details))
        else:
            log.error(
                '{} celery was disabled but the task={} '
                'did not return anything'.format(
                    label,
                    response))
        # end of if response
    else:
        task_res = prepare_pricing_dataset.delay(
            work_dict=work_dict)
        rec = {
            'task_id': task_res
        }
        response = build_result.build_result(
            status=ae_consts.SUCCESS,
            err=None,
            rec=rec)
    # if celery enabled

    if response:
        if ae_consts.ev('DEBUG_RESULTS', '0') == '1':
            log.info(
                'run_prepare_pricing_dataset - {} - done '
                'status={} err={} rec={}'.format(
                    label,
                    ae_consts.get_status(response['status']),
                    response['err'],
                    response['rec']))
        else:
            log.info(
                'run_prepare_pricing_dataset - {} - done '
                'status={} err={}'.format(
                    label,
                    ae_consts.get_status(response['status']),
                    response['err']))
    else:
        log.info(
            'run_prepare_pricing_dataset - {} - done '
            'no response'.format(
                label))
    # end of if/else response

    return response
# end of run_prepare_pricing_dataset



"""
Helper for creating a sell order
"""

import analysis_engine.consts as ae_consts
import analysis_engine.utils as ae_utils
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def build_sell_order(
        ticker,
        num_owned,
        close,
        balance,
        commission,
        date,
        details,
        use_key,
        minute=None,
        shares=None,
        version=1,
        auto_fill=True,
        is_live_trading=False,
        backtest_shares_default=10,
        reason=None):
    """build_sell_order

    Create an algorithm sell order as a dictionary

    :param ticker: ticker
    :param num_owned: integer current owned
        number of shares for this asset
    :param close: float closing price of the asset
    :param balance: float amount of available capital
    :param commission: float for commission costs
    :param date: string trade date for that row usually
        ``COMMON_DATE_FORMAT`` (``YYYY-MM-DD``)
    :param minute: optional - string with the minute that the
        order was placed. format is
        ``COMMON_TICK_DATE_FORMAT`` (``YYYY-MM-DD HH:MM:SS``)
    :param details: dictionary for full row of values to review
        all sells after the algorithm finishes.
        (usually ``row.to_json()``)
    :param use_key: string for redis and s3 publishing of the algorithm
        result dictionary as a json-serialized dictionary
    :param shares: optional - integer number of shares to sell
        if None sell all ``num_owned`` shares at the ``close``.
    :param version: optional - version tracking integer
    :param auto_fill: optional - bool for not assuming the trade
        filled (default ``True``)
    :param is_live_trading: optional - bool for filling trades
        for live trading or for backtest tuning filled
        (default ``False`` which is backtest mode)
    :param backtest_shares_default: optional - integer for
        simulating shares during a backtest even if there
        are not enough funds
        (default ``10``)
    :param reason: optional - string for recording why the algo
        decided to sell for review after the algorithm finishes
    """
    status = ae_consts.TRADE_OPEN
    s3_bucket_name = ae_consts.ALGO_SELLS_S3_BUCKET_NAME
    s3_key = use_key
    redis_key = use_key
    s3_enabled = True
    redis_enabled = True

    cost_of_trade = None
    sell_price = 0.0
    new_shares = num_owned
    new_balance = balance
    created_date = None

    tradable_funds = balance - commission

    if num_owned == 0:
        status = ae_consts.TRADE_NO_SHARES_TO_SELL
    elif close > 0.1 and tradable_funds > 10.0:
        cost_of_trade = commission
        if shares:
            if shares > num_owned:
                shares = num_owned
        else:
            shares = num_owned
        sell_price = ae_consts.to_f(
            val=(shares * close) + commission)
        if cost_of_trade > balance:
            status = ae_consts.TRADE_NOT_ENOUGH_FUNDS
        else:
            created_date = ae_utils.utc_now_str()
            if auto_fill:
                new_shares = num_owned - shares
                new_balance = ae_consts.to_f(balance + sell_price)
                status = ae_consts.TRADE_FILLED
            else:
                new_shares = shares
                new_balance = balance
    else:
        status = ae_consts.TRADE_NOT_ENOUGH_FUNDS

    order_dict = {
        'ticker': ticker,
        'status': status,
        'balance': new_balance,
        'shares': new_shares,
        'sell_price': sell_price,
        'prev_balance': balance,
        'prev_shares': num_owned,
        'close': close,
        'details': details,
        'reason': reason,
        'date': date,
        'minute': minute,
        'created': created_date,
        's3_bucket': s3_bucket_name,
        's3_key': s3_key,
        'redis_key': redis_key,
        's3_enabled': s3_enabled,
        'redis_enabled': redis_enabled,
        'version': version
    }

    use_date = minute
    if not use_date:
        use_date = date

    log.debug(
        '{} {} sell {} order={}'.format(
            ticker,
            use_date,
            ae_consts.get_status(status=order_dict['status']),
            ae_consts.ppj(order_dict)))

    return order_dict
# end of build_sell_order



"""
Fetch data from Tradier:
https://developer.tradier.com/getting_started
"""

import json
import copy
import pandas as pd
import analysis_engine.consts as ae_consts
import analysis_engine.td.consts as td_consts
import analysis_engine.options_dates as opt_dates
import analysis_engine.td.fetch_api as td_fetch
import analysis_engine.td.extract_df_from_redis as td_extract
import spylunking.log.setup_logging as log_utils

log = log_utils.build_colorized_logger(name=__name__)


def fetch_data(
        work_dict,
        fetch_type=None):
    """fetch_data

    Factory method for fetching data from
    TD using an enum or string alias. Returns
    a pandas ``DataFrame`` and only supports
    one ticker at a time.

    Supported enums from: ``analysis_engine.td.consts``

    ::

        fetch_type = FETCH_TD_CALLS
        fetch_type = FETCH_TD_PUTS

    Supported ``work_dict['ft_type']`` string values:

    ::

        work_dict['ft_type'] = 'tdcalls'
        work_dict['ft_type'] = 'tdputs'

    :param work_dict: dictionary of args for the Tradier api
    :param fetch_type: optional - name or enum of the fetcher to create
                       can also be a lower case string
                       in work_dict['ft_type']
    """
    use_fetch_name = None
    ticker = work_dict.get(
        'ticker',
        None)
    if not fetch_type:
        fetch_type = work_dict.get(
            'ft_type',
            None)
    if fetch_type:
        use_fetch_name = str(fetch_type).lower()

    if 'exp_date' not in work_dict:
        work_dict['exp_date'] = opt_dates.option_expiration().strftime(
            ae_consts.COMMON_DATE_FORMAT)

    log.debug(
        'name={} type={} args={}'.format(
            use_fetch_name,
            fetch_type,
            work_dict))

    status_df = ae_consts.NOT_SET
    df = pd.DataFrame([{}])

    if (
            use_fetch_name == 'tdcalls' or
            fetch_type == td_consts.FETCH_TD_CALLS):
        status_df, fetch_df = td_fetch.fetch_calls(
            work_dict=work_dict)

        if status_df == ae_consts.SUCCESS:
            log.debug(
                'call - merge df')
            work_copy = copy.deepcopy(
                work_dict)
            work_copy['ft_type'] = td_consts.FETCH_TD_CALLS
            work_copy['fd_type'] = 'tdcalls'
            if 'tdcalls' in work_dict:
                work_copy['redis_key'] = work_dict['tdcalls']
                work_copy['s3_key'] = '{}.json'.format(
                    work_dict['tdcalls'])
            else:
                work_copy['redis_key'] = (
                    '{}_tdcalls'.format(
                        work_dict['redis_key']))
                work_copy['s3_key'] = (
                    '{}_tdcalls'.format(
                        work_dict['s3_key']))
            ext_status, ext_df = \
                td_extract.extract_option_calls_dataset(
                    work_dict=work_copy)
            if ext_status == ae_consts.SUCCESS and len(ext_df.index) > 0:
                log.debug(
                    'call - merging fetch={} with ext={}'.format(
                        len(fetch_df.index),
                        len(ext_df.index)))
                """
                for testing compression:
                """
                """
                import sys
                print(ext_df['date'])
                print(ext_df['ask_date'])
                print(ext_df['bid_date'])
                print(ext_df['trade_date'])
                sys.exit(1)
                """
                extracted_records = json.loads(ext_df.to_json(
                    orient='records'))
                fetched_records = json.loads(fetch_df.to_json(
                    orient='records'))
                new_records = []
                dates_by_strike_dict = {}
                for ex_row in extracted_records:
                    date_strike_name = (
                        '{}_{}'.format(
                            ex_row['created'],
                            ex_row['strike']))
                    if date_strike_name not in dates_by_strike_dict:
                        new_node = {}
                        for c in td_consts.TD_OPTION_COLUMNS:
                            if c in ex_row:
                                new_node[c] = ex_row[c]
                        # end of for all columns to copy over
                        new_node.pop('index', None)
                        new_node.pop('level_0', None)
                        new_records.append(new_node)
                        dates_by_strike_dict[date_strike_name] = True
                # build extracted records

                for ft_row in fetched_records:
                    date_strike_name = (
                        '{}_{}'.format(
                            ft_row['created'],
                            ft_row['strike']))
                    try:
                        if date_strike_name not in dates_by_strike_dict:
                            new_node = {}
                            for c in td_consts.TD_OPTION_COLUMNS:
                                if c in ft_row:
                                    new_node[c] = ft_row[c]
                            # end of for all columns to copy over
                            new_node.pop('index', None)
                            new_node.pop('level_0', None)
                            new_records.append(new_node)
                            dates_by_strike_dict[date_strike_name] = True
                        else:
                            log.error(
                                'already have {} call - date={} '
                                'strike={}'.format(
                                    ticker,
                                    ft_row['created'],
                                    ft_row['strike']))
                    except Exception as p:
                        log.critical(
                            'failed fetching call with ex={}'.format(
                                p))
                        return ae_consts.ERR, None
                    # end of adding fetched records after the extracted

                df = pd.DataFrame(new_records)
                df.sort_values(
                    by=[
                        'date',
                        'strike'
                    ],
                    ascending=True)
                log.debug(
                    'call - merged={}'.format(
                        len(df.index)))
            else:
                df = fetch_df.sort_values(
                    by=[
                        'date',
                        'strike'
                    ],
                    ascending=True)
        else:
            log.warn(
                f'{ticker} - no data found for calls')
        # if able to merge fetch + last for today
    elif (
            use_fetch_name == 'tdputs' or
            fetch_type == td_consts.FETCH_TD_PUTS):
        status_df, fetch_df = td_fetch.fetch_puts(
            work_dict=work_dict)
        if status_df == ae_consts.SUCCESS:
            log.debug(
                'put - merge df')
            work_copy = copy.deepcopy(
                work_dict)
            work_copy['ft_type'] = td_consts.FETCH_TD_PUTS
            work_copy['fd_type'] = 'tdputs'
            if 'tdputs' in work_dict:
                work_copy['redis_key'] = work_dict['tdputs']
                work_copy['s3_key'] = '{}.json'.format(
                    work_dict['tdputs'])
            else:
                work_copy['redis_key'] = (
                    '{}_tdputs'.format(
                        work_dict['redis_key']))
                work_copy['s3_key'] = (
                    '{}_tdputs'.format(
                        work_dict['s3_key']))
            ext_status, ext_df = \
                td_extract.extract_option_puts_dataset(
                    work_dict=work_copy)
            if ext_status == ae_consts.SUCCESS and len(ext_df.index) > 0:
                log.debug(
                    'put - merging fetch={} with ext={}'.format(
                        len(fetch_df.index),
                        len(ext_df.index)))
                """
                for testing compression:
                """
                """
                import sys
                print(ext_df['date'])
                sys.exit(1)
                """
                extracted_records = json.loads(ext_df.to_json(
                    orient='records'))
                fetched_records = json.loads(fetch_df.to_json(
                    orient='records'))
                new_records = []
                dates_by_strike_dict = {}
                for ex_row in extracted_records:
                    date_strike_name = (
                        '{}_{}'.format(
                            ex_row['created'],
                            ex_row['strike']))
                    if date_strike_name not in dates_by_strike_dict:
                        new_node = {}
                        for c in td_consts.TD_OPTION_COLUMNS:
                            if c in ex_row:
                                new_node[c] = ex_row[c]
                        # end of for all columns to copy over
                        new_node.pop('index', None)
                        new_node.pop('level_0', None)
                        new_records.append(new_node)
                        dates_by_strike_dict[date_strike_name] = True
                # build extracted records

                for ft_row in fetched_records:
                    date_strike_name = (
                        '{}_{}'.format(
                            ft_row['created'],
                            ft_row['strike']))
                    try:
                        if date_strike_name not in dates_by_strike_dict:
                            new_node = {}
                            for c in td_consts.TD_OPTION_COLUMNS:
                                if c in ft_row:
                                    new_node[c] = ft_row[c]
                            # end of for all columns to copy over
                            new_node.pop('index', None)
                            new_node.pop('level_0', None)
                            new_records.append(new_node)
                            dates_by_strike_dict[date_strike_name] = True
                        else:
                            log.error(
                                'already have {} put - date={} '
                                'strike={}'.format(
                                    ticker,
                                    ft_row['created'],
                                    ft_row['strike']))
                    except Exception as p:
                        log.critical(
                            'failed fetching puts with ex={}'.format(
                                p))
                        return ae_consts.ERR, None
                # end of adding fetched records after the extracted

                df = pd.DataFrame(new_records)
                df.sort_values(
                    by=[
                        'date',
                        'strike'
                    ],
                    ascending=True)
                log.debug(
                    'put - merged={}'.format(
                        len(df.index)))
            else:
                df = fetch_df.sort_values(
                    by=[
                        'date',
                        'strike'
                    ],
                    ascending=True)
        else:
            log.warn(
                f'{ticker} - no data found for puts')
        # if able to merge fetch + last for today
    else:
        log.error(
            f'label={work_dict.get("label", None)} - '
            f'unsupported fetch_data('
            f'work_dict={work_dict}, '
            f'fetch_type={fetch_type}'
            f')')
        raise NotImplemented
    # end of supported fetchers

    return status_df, df
# end of fetch_data



#!/usr/bin/env python

import os
import sys
import datetime
import json
import argparse
import pandas as pd
import analysis_engine.consts as ae_consts
import analysis_engine.plot_trading_history as plot_trading_history
import analysis_engine.load_history_dataset as load_history_utils
import analysis_engine.run_custom_algo as run_custom_algo
import spylunking.log.setup_logging as log_utils


log = log_utils.build_colorized_logger(
    name='ae',
    log_config_path=ae_consts.LOG_CONFIG_PATH)


def start_algo():
    """start_algo

    Run a derived algorithm with an algorithm config dictionary

    :param config_dict: algorithm config dictionary
    """

    log.debug('start - ae')

    parser = argparse.ArgumentParser(
        description=(
            'stock analysis tool'))
    parser.add_argument(
        '-t',
        help=(
            'ticker'),
        required=False,
        dest='ticker')
    parser.add_argument(
        '-e',
        help=(
            'file path to extract an '
            'algorithm-ready datasets from redis'),
        required=False,
        dest='algo_extract_loc')
    parser.add_argument(
        '-l',
        help=(
            'show dataset in this file'),
        required=False,
        dest='show_from_file')
    parser.add_argument(
        '-H',
        help=(
            'show trading history dataset in this file'),
        required=False,
        dest='show_history_from_file')
    parser.add_argument(
        '-E',
        help=(
            'show trading performance report dataset in this file'),
        required=False,
        dest='show_report_from_file')
    parser.add_argument(
        '-L',
        help=(
            'restore an algorithm-ready dataset file back into redis'),
        required=False,
        dest='restore_algo_file')
    parser.add_argument(
        '-f',
        help=(
            'save the trading history dataframe '
            'to this file'),
        required=False,
        dest='history_json_file')
    parser.add_argument(
        '-J',
        help=(
            'plot action - after preparing you can use: '
            '-J show to open the image (good for debugging)'),
        required=False,
        dest='plot_action')
    parser.add_argument(
        '-b',
        help=(
            'run a backtest using the dataset in '
            'a file path/s3 key/redis key formats: '
            'file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
            's3://algoready/SPY-latest.json or '
            'redis:SPY-latest'),
        required=False,
        dest='backtest_loc')
    parser.add_argument(
        '-B',
        help=(
            'optional - broker url for Celery'),
        required=False,
        dest='broker_url')
    parser.add_argument(
        '-C',
        help=(
            'optional - broker url for Celery'),
        required=False,
        dest='backend_url')
    parser.add_argument(
        '-w',
        help=(
            'optional - flag for publishing an algorithm job '
            'using Celery to the ae workers'),
        required=False,
        dest='run_on_engine',
        action='store_true')
    parser.add_argument(
        '-k',
        help=(
            'optional - s3 access key'),
        required=False,
        dest='s3_access_key')
    parser.add_argument(
        '-K',
        help=(
            'optional - s3 secret key'),
        required=False,
        dest='s3_secret_key')
    parser.add_argument(
        '-a',
        help=(
            'optional - s3 address format: <host:port>'),
        required=False,
        dest='s3_address')
    parser.add_argument(
        '-Z',
        help=(
            'optional - s3 secure: default False'),
        required=False,
        dest='s3_secure')
    parser.add_argument(
        '-s',
        help=(
            'optional - start date: YYYY-MM-DD'),
        required=False,
        dest='start_date')
    parser.add_argument(
        '-n',
        help=(
            'optional - end date: YYYY-MM-DD'),
        required=False,
        dest='end_date')
    parser.add_argument(
        '-u',
        help=(
            'optional - s3 bucket name'),
        required=False,
        dest='s3_bucket_name')
    parser.add_argument(
        '-G',
        help=(
            'optional - s3 region name'),
        required=False,
        dest='s3_region_name')
    parser.add_argument(
        '-g',
        help=(
            'Path to a custom algorithm module file '
            'on disk. This module must have a single '
            'class that inherits from: '
            'https://github.com/AlgoTraders/stock-analysis-engine/'
            'blob/master/'
            'analysis_engine/algo.py Additionally you '
            'can find the Example-Minute-Algorithm here: '
            'https://github.com/AlgoTraders/stock-analysis-engine/'
            'blob/master/analysis_engine/mocks/'
            'example_algo_minute.py'),
        required=False,
        dest='run_algo_in_file')
    parser.add_argument(
        '-p',
        help=(
            'optional - s3 bucket/file for trading history'),
        required=False,
        dest='algo_history_loc')
    parser.add_argument(
        '-o',
        help=(
            'optional - s3 bucket/file for trading performance report'),
        required=False,
        dest='algo_report_loc')
    parser.add_argument(
        '-r',
        help=(
            'optional - redis_address format: <host:port>'),
        required=False,
        dest='redis_address')
    parser.add_argument(
        '-R',
        help=(
            'optional - redis and s3 key name'),
        required=False,
        dest='keyname')
    parser.add_argument(
        '-m',
        help=(
            'optional - redis database number (0 by default)'),
        required=False,
        dest='redis_db')
    parser.add_argument(
        '-x',
        help=(
            'optional - redis expiration in seconds'),
        required=False,
        dest='redis_expire')
    parser.add_argument(
        '-c',
        help=(
            'optional - algorithm config_file path for setting '
            'up internal algorithm trading strategies and '
            'indicators'),
        required=False,
        dest='config_file')
    parser.add_argument(
        '-v',
        help=(
            'set the Algorithm to verbose logging'),
        required=False,
        dest='verbose_algo',
        action='store_true')
    parser.add_argument(
        '-P',
        help=(
            'set the Algorithm\'s IndicatorProcessor to verbose logging'),
        required=False,
        dest='verbose_processor',
        action='store_true')
    parser.add_argument(
        '-I',
        help=(
            'set all Algorithm\'s Indicators to verbose logging '
            '(note indivdual indicators support a \'verbose\' key '
            'that can be set to True to debug just one '
            'indicator)'),
        required=False,
        dest='verbose_indicators',
        action='store_true')
    parser.add_argument(
        '-V',
        help=(
            'inspect the datasets an algorithm is processing - this'
            'will slow down processing to show debugging'),
        required=False,
        dest='inspect_datasets',
        action='store_true')
    parser.add_argument(
        '-j',
        help=(
            'run the algorithm on just this specific date in the datasets '
            '- specify the date in a format: YYYY-MM-DD like: 2018-11-29'),
        required=False,
        dest='run_this_date')
    parser.add_argument(
        '-d',
        help=(
            'debug'),
        required=False,
        dest='debug',
        action='store_true')
    args = parser.parse_args()

    ticker = None
    use_config_file = '/opt/sa/cfg/default_algo.json'
    algo_mod_path = '/opt/sa/analysis_engine/algo.py'
    use_balance = 10000.0
    use_commission = 6.0
    use_start_date = None
    use_end_date = None
    debug = False
    verbose_algo = None
    verbose_processor = None
    verbose_indicators = None
    inspect_datasets = None
    history_json_file = None
    run_this_date = None
    algo_obj = None
    algo_history_loc = None
    algo_report_loc = None
    algo_extract_loc = None
    backtest_loc = None

    ssl_options = ae_consts.SSL_OPTIONS
    transport_options = ae_consts.TRANSPORT_OPTIONS
    broker_url = ae_consts.WORKER_BROKER_URL
    backend_url = ae_consts.WORKER_BACKEND_URL
    path_to_config_module = ae_consts.WORKER_CELERY_CONFIG_MODULE
    include_tasks = ae_consts.INCLUDE_TASKS
    load_from_s3_bucket = None
    load_from_s3_key = None
    load_from_redis_key = None
    load_from_file = None
    load_compress = True
    load_publish = True
    load_config = None
    report_redis_key = None
    report_s3_bucket = None
    report_s3_key = None
    report_file = None
    report_compress = True
    report_publish = False
    report_config = None
    history_redis_key = None
    history_s3_bucket = None
    history_s3_key = None
    history_file = None
    history_compress = True
    history_publish = True
    history_config = None
    extract_redis_key = None
    extract_s3_bucket = None
    extract_s3_key = None
    extract_file = None
    extract_save_dir = None
    extract_compress = False
    extract_publish = False
    extract_config = None
    s3_enabled = True
    s3_access_key = ae_consts.S3_ACCESS_KEY
    s3_secret_key = ae_consts.S3_SECRET_KEY
    s3_region_name = ae_consts.S3_REGION_NAME
    s3_address = ae_consts.S3_ADDRESS
    s3_bucket_name = ae_consts.S3_BUCKET
    s3_key = None
    s3_secure = ae_consts.S3_SECURE
    redis_enabled = True
    redis_address = ae_consts.REDIS_ADDRESS
    redis_key = None
    redis_password = ae_consts.REDIS_PASSWORD
    redis_db = ae_consts.REDIS_DB
    redis_expire = ae_consts.REDIS_EXPIRE
    redis_serializer = 'json'
    redis_encoding = 'utf-8'
    publish_to_s3 = True
    publish_to_redis = False
    publish_to_slack = False
    slack_enabled = False
    slack_code_block = False
    slack_full_width = False

    dataset_type = ae_consts.SA_DATASET_TYPE_ALGO_READY
    serialize_datasets = ae_consts.DEFAULT_SERIALIZED_DATASETS
    compress = False
    encoding = 'utf-8'
    debug = False
    run_on_engine = False

    auto_fill = True
    timeseries = 'minute'
    trade_strategy = 'count'

    if args.s3_access_key:
        s3_access_key = args.s3_access_key
    if args.s3_secret_key:
        s3_secret_key = args.s3_secret_key
    if args.s3_region_name:
        s3_region_name = args.s3_region_name
    if args.s3_address:
        s3_address = args.s3_address
    if args.s3_secure:
        s3_secure = args.s3_secure
    if args.redis_address:
        redis_address = args.redis_address
    if args.redis_db:
        redis_db = args.redis_db
    if args.redis_expire:
        redis_expire = args.redis_expire
    if args.history_json_file:
        history_json_file = args.history_json_file
    if args.ticker:
        ticker = args.ticker.upper()
    if args.debug:
        debug = True
    if args.verbose_algo:
        verbose_algo = True
    if args.verbose_processor:
        verbose_processor = True
    if args.verbose_indicators:
        verbose_indicators = True
    if args.inspect_datasets:
        inspect_datasets = True
    if args.run_this_date:
        run_this_date = args.run_this_date
    if args.start_date:
        try:
            use_start_date = '{} 00:00:00'.format(
                str(args.start_date))
            datetime.datetime.strptime(
                args.start_date,
                ae_consts.COMMON_DATE_FORMAT)
        except Exception as e:
            msg = (
                'please use a start date formatted as: {}'
                '\n'
                'error was: {}'.format(
                    ae_consts.COMMON_DATE_FORMAT,
                    e))
            log.error(msg)
            sys.exit(1)
        # end of testing for a valid date
    # end of args.start_date
    if args.end_date:
        try:
            use_end_date = '{} 00:00:00'.format(
                str(args.end_date))
            datetime.datetime.strptime(
                args.end_date,
                ae_consts.COMMON_DATE_FORMAT)
        except Exception as e:
            msg = (
                'please use an end date formatted as: {}'
                '\n'
                'error was: {}'.format(
                    ae_consts.COMMON_DATE_FORMAT,
                    e))
            log.error(msg)
            sys.exit(1)
        # end of testing for a valid date
    # end of args.end_date
    config_dict = None
    if args.run_algo_in_file:
        if not os.path.exists(args.run_algo_in_file):
            log.error(
                'missing algorithm module file: {}'.format(
                    args.run_algo_in_file))
            sys.exit(1)
        algo_mod_path = args.run_algo_in_file
    if args.config_file:
        use_config_file = args.config_file

    if not os.path.exists(use_config_file):
        log.error(
            'Failed: unable to find config file: -c {}'.format(
                use_config_file))
        sys.exit(1)
    config_dict = json.loads(open(use_config_file).read())
    algo_mod_path = config_dict.get(
        'algo_path',
        algo_mod_path)
    if not os.path.exists(algo_mod_path):
        log.error(
            'missing algorithm module file from config: {}'.format(
                algo_mod_path))
        sys.exit(1)

    """
    Finalize the algo config
    """
    if config_dict:
        use_balance = float(config_dict.get(
            'balance',
            use_balance))
        use_commission = float(config_dict.get(
            'commission',
            use_commission))
        ticker = str(config_dict.get(
            'ticker',
            ticker)).upper()

        config_dict['ticker'] = ticker
        config_dict['balance'] = use_balance
        config_dict['commission'] = use_commission
    else:
        if not ticker:
            ticker = str(config_dict.get(
                'ticker',
                ae_consts.TICKER)).upper()
    if not ticker:
        log.error(
            'usage error: please set a ticker with -t <TICKER>')
        sys.exit(1)

    if verbose_algo:
        config_dict['verbose'] = verbose_algo
    if verbose_processor:
        config_dict['verbose_processor'] = verbose_processor
    if verbose_indicators:
        config_dict['verbose_indicators'] = verbose_indicators
    if inspect_datasets:
        config_dict['inspect_datasets'] = inspect_datasets
    if run_this_date:
        config_dict['run_this_date'] = run_this_date

    log.info(
        'starting')

    """
    Run a custom algo module from disk
    """
    if algo_mod_path:

        if args.backtest_loc:
            backtest_loc = args.backtest_loc
            if ('file:/' not in backtest_loc and
                    's3://' not in backtest_loc and
                    'redis://' not in backtest_loc):
                log.error(
                    'invalid -b <backtest dataset file> specified. '
                    '{} '
                    'please use either: '
                    '-b file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
                    '-b s3://algoready/SPY-latest.json or '
                    '-b redis://SPY-latest'.format(
                        backtest_loc))
                sys.exit(1)
            if 's3://' in backtest_loc:
                load_from_s3_bucket = backtest_loc.split('/')[-2]
                load_from_s3_key = backtest_loc.split('/')[-1]
            elif 'redis://' in backtest_loc:
                load_from_redis_key = backtest_loc.split('/')[-1]
            elif 'file:/' in backtest_loc:
                load_from_file = backtest_loc.split(':')[-1]
            load_publish = True
        # end of parsing supported transport - loading an algo-ready

        if args.algo_history_loc:
            algo_history_loc = args.algo_history_loc
            if ('file:/' not in algo_history_loc and
                    's3://' not in algo_history_loc and
                    'redis://' not in algo_history_loc):
                log.error(
                    'invalid -p <backtest dataset file> specified. '
                    '{} '
                    'please use either: '
                    '-p file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
                    '-p s3://algoready/SPY-latest.json or '
                    '-p redis://SPY-latest'.format(
                        algo_history_loc))
                sys.exit(1)
            if 's3://' in algo_history_loc:
                history_s3_bucket = algo_history_loc.split('/')[-2]
                history_s3_key = algo_history_loc.split('/')[-1]
            elif 'redis://' in algo_history_loc:
                history_redis_key = algo_history_loc.split('/')[-1]
            elif 'file:/' in algo_history_loc:
                history_file = algo_history_loc.split(':')[-1]
            history_publish = True
        # end of parsing supported transport - trading history

        if args.algo_report_loc:
            algo_report_loc = args.algo_report_loc
            if ('file:/' not in algo_report_loc and
                    's3://' not in algo_report_loc and
                    'redis://' not in algo_report_loc):
                log.error(
                    'invalid -o <backtest dataset file> specified. '
                    '{} '
                    'please use either: '
                    '-o file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
                    '-o s3://algoready/SPY-latest.json or '
                    '-o redis://SPY-latest'.format(
                        algo_report_loc))
                sys.exit(1)
            if 's3://' in algo_report_loc:
                report_s3_bucket = algo_report_loc.split('/')[-2]
                report_s3_key = algo_report_loc.split('/')[-1]
            elif 'redis://' in algo_report_loc:
                report_redis_key = algo_report_loc.split('/')[-1]
            elif 'file:/' in algo_report_loc:
                report_file = algo_report_loc.split(':')[-1]
            report_publish = True
        # end of parsing supported transport - trading performance report

        if args.algo_extract_loc:
            algo_extract_loc = args.algo_extract_loc
            if ('file:/' not in algo_extract_loc and
                    's3://' not in algo_extract_loc and
                    'redis://' not in algo_extract_loc):
                log.error(
                    'invalid -e <backtest dataset file> specified. '
                    '{} '
                    'please use either: '
                    '-e file:/opt/sa/tests/datasets/algo/SPY-latest.json or '
                    '-e s3://algoready/SPY-latest.json or '
                    '-e redis://SPY-latest'.format(
                        algo_extract_loc))
                sys.exit(1)
            if 's3://' in algo_extract_loc:
                extract_s3_bucket = algo_extract_loc.split('/')[-2]
                extract_s3_key = algo_extract_loc.split('/')[-1]
            elif 'redis://' in algo_extract_loc:
                extract_redis_key = algo_extract_loc.split('/')[-1]
            elif 'file:/' in algo_extract_loc:
                extract_file = algo_extract_loc.split(':')[-1]
            extract_publish = True
        # end of parsing supported transport - extract algorithm-ready

        if args.run_on_engine:
            run_on_engine = True
            if verbose_algo:
                log.info('starting algo on the engine')

        use_name = config_dict.get(
            'name',
            'missing-algo-name')
        auto_fill = config_dict.get(
            'auto_fill',
            auto_fill)
        timeseries = config_dict.get(
            'timeseries',
            timeseries)
        trade_strategy = config_dict.get(
            'trade_strategy',
            trade_strategy)

        algo_res = run_custom_algo.run_custom_algo(
            mod_path=algo_mod_path,
            ticker=config_dict['ticker'],
            balance=config_dict['balance'],
            commission=config_dict['commission'],
            name=use_name,
            start_date=use_start_date,
            end_date=use_end_date,
            auto_fill=auto_fill,
            config_dict=config_dict,
            load_from_s3_bucket=load_from_s3_bucket,
            load_from_s3_key=load_from_s3_key,
            load_from_redis_key=load_from_redis_key,
            load_from_file=load_from_file,
            load_compress=load_compress,
            load_publish=load_publish,
            load_config=load_config,
            report_redis_key=report_redis_key,
            report_s3_bucket=report_s3_bucket,
            report_s3_key=report_s3_key,
            report_file=report_file,
            report_compress=report_compress,
            report_publish=report_publish,
            report_config=report_config,
            history_redis_key=history_redis_key,
            history_s3_bucket=history_s3_bucket,
            history_s3_key=history_s3_key,
            history_file=history_file,
            history_compress=history_compress,
            history_publish=history_publish,
            history_config=history_config,
            extract_redis_key=extract_redis_key,
            extract_s3_bucket=extract_s3_bucket,
            extract_s3_key=extract_s3_key,
            extract_file=extract_file,
            extract_save_dir=extract_save_dir,
            extract_compress=extract_compress,
            extract_publish=extract_publish,
            extract_config=extract_config,
            publish_to_slack=publish_to_slack,
            publish_to_s3=publish_to_s3,
            publish_to_redis=publish_to_redis,
            dataset_type=dataset_type,
            serialize_datasets=serialize_datasets,
            compress=compress,
            encoding=encoding,
            redis_enabled=redis_enabled,
            redis_key=redis_key,
            redis_address=redis_address,
            redis_db=redis_db,
            redis_password=redis_password,
            redis_expire=redis_expire,
            redis_serializer=redis_serializer,
            redis_encoding=redis_encoding,
            s3_enabled=s3_enabled,
            s3_key=s3_key,
            s3_address=s3_address,
            s3_bucket=s3_bucket_name,
            s3_access_key=s3_access_key,
            s3_secret_key=s3_secret_key,
            s3_region_name=s3_region_name,
            s3_secure=s3_secure,
            slack_enabled=slack_enabled,
            slack_code_block=slack_code_block,
            slack_full_width=slack_full_width,
            dataset_publish_extract=extract_publish,
            dataset_publish_history=history_publish,
            dataset_publish_report=report_publish,
            run_on_engine=run_on_engine,
            auth_url=broker_url,
            backend_url=backend_url,
            include_tasks=include_tasks,
            ssl_options=ssl_options,
            transport_options=transport_options,
            path_to_config_module=path_to_config_module,
            timeseries=timeseries,
            trade_strategy=trade_strategy,
            verbose=verbose_algo)

        show_label = 'algo.name={}'.format(
            use_name)
        show_extract = '{}'.format(
            algo_extract_loc)
        show_history = '{}'.format(
            algo_history_loc)
        show_report = '{}'.format(
            algo_report_loc)
        base_label = (
            'load={} '
            'extract={} '
            'history={} '
            'report={}'.format(
                args.run_algo_in_file,
                show_extract,
                show_history,
                show_report))

        task_id = None
        if run_on_engine:
            task_id = algo_res.get('rec', {}).get('task_id', None)
        if task_id:
            log.info(
                'waiting on task_id={} to finish'.format(
                    task_id))
            res = task_id.get()
            history_config = res.get(
                'algo_req', {}).get(
                    'history_config', None)
            s3_bucket = history_config.get('s3_bucket', None)
            s3_key = history_config.get('s3_key', None)
            load_res = load_history_utils.load_history_dataset(
                s3_bucket=s3_bucket,
                s3_key=s3_key)

            history_df = load_res[ticker]
            log.info(
                'plotting history df with columns: {}'.format(
                    history_df.columns.values))

            history_df['date'] = pd.to_datetime(
                history_df['date'])
            history_df['minute'] = pd.to_datetime(
                history_df['minute'])
            ticker = history_df['ticker'].iloc[0]

            first_date = history_df['date'].iloc[0]
            end_date = history_df['date'].iloc[-1]
            title = (
                'Trading History {}\n'
                'Backtest dates from {} to {}'.format(
                    ticker,
                    first_date,
                    end_date))
            use_xcol = 'date'
            use_as_date_format = '%d\n%b'
            use_minute = False
            if 'minute' in history_df:
                found_valid_minute = history_df['minute'].iloc[0]
                if found_valid_minute:
                    use_minute = True

            if use_minute:
                use_xcol = 'minute'
                use_as_date_format = '%d %H:%M:%S\n%b'
            xlabel = 'Dates vs Algo values'
            ylabel = 'Algo values'
            df_filter = (history_df['close'] > 1.00)

            # set default columns:
            blue = None
            green = None
            orange = None

            red = 'balance'
            blue = 'close'

            plot_trading_history.plot_trading_history(
                title=title,
                df=history_df,
                red=red,
                blue=blue,
                green=green,
                orange=orange,
                date_col=use_xcol,
                date_format=use_as_date_format,
                xlabel=xlabel,
                ylabel=ylabel,
                df_filter=df_filter,
                show_plot=True,
                dropna_for_all=True)

            return
        else:
            algo_obj = algo_res.get(
                'algo',
                None)
        # end of getting the algo results from s3

        if not algo_obj:
            log.error(
                '{} - did not create algorithm object'.format(
                    show_label))
            sys.exit(1)

        if not run_on_engine:
            algo_trade_history_recs = algo_res['rec'].get(
                'history',
                [])
            show_label = (
                '{} algo.name={} {} trade_history_len={}'.format(
                    ticker,
                    use_name,
                    base_label,
                    len(algo_trade_history_recs)))
        if args.debug:
            log.info(
                'algo_res={}'.format(
                    algo_res))
            if algo_res['status'] == ae_consts.SUCCESS:
                log.info(
                    '{} - done running {}'.format(
                        ae_consts.get_status(
                            status=algo_res['status']),
                        show_label))
            else:
                log.error(
                    '{} - done running {}'.format(
                        ae_consts.get_status(
                            status=algo_res['status']),
                        show_label))
        else:
            if algo_res['status'] == ae_consts.SUCCESS:
                log.info(
                    '{} - done running {}'.format(
                        ae_consts.get_status(
                            status=algo_res['status']),
                        show_label))
            else:
                log.error(
                    'run_custom_algo returned error: {}'.format(
                        algo_res['err']))
                sys.exit(1)
        # end of running the custom algo handler

    else:
        log.error(
            f'missing an algorithm mod path: {algo_mod_path}')
        sys.exit(1)
    # end if running a custom algorithm module

    if algo_obj:

        trading_history_dict = algo_obj.get_history_dataset()
        history_df = trading_history_dict[ticker]
        if not hasattr(history_df, 'to_json'):
            return

        if history_json_file:
            log.info(
                'saving history to: {}'.format(
                    history_json_file))
            history_df.to_json(
                history_json_file,
                orient='records',
                date_format='iso')

        log.info('plotting history')

        first_date = history_df['date'].iloc[0]
        end_date = history_df['date'].iloc[-1]
        title = (
            'Trading History {} for Algo {}\n'
            'Backtest dates from {} to {}'.format(
                ticker,
                trading_history_dict['algo_name'],
                first_date,
                end_date))
        use_xcol = 'date'
        use_as_date_format = '%d\n%b'
        if config_dict['timeseries'] == 'minute':
            use_xcol = 'minute'
            use_as_date_format = '%d %H:%M:%S\n%b'
        xlabel = 'Dates vs {} values'.format(
            trading_history_dict['algo_name'])
        ylabel = 'Algo {}\nvalues'.format(
            trading_history_dict['algo_name'])
        df_filter = (history_df['close'] > 0.01)

        # set default hloc columns:
        blue = None
        green = None
        orange = None

        red = 'close'
        blue = 'balance'

        if debug:
            for i, r in history_df.iterrows():
                log.info('{} - {}'.format(
                    r['minute'],
                    r['close']))

        plot_trading_history.plot_trading_history(
            title=title,
            df=history_df,
            red=red,
            blue=blue,
            green=green,
            orange=orange,
            date_col=use_xcol,
            date_format=use_as_date_format,
            xlabel=xlabel,
            ylabel=ylabel,
            df_filter=df_filter,
            show_plot=True,
            dropna_for_all=True)

# end of start_algo


if __name__ == '__main__':
    start_algo()
